# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np

import galsim
from galsim.zernike import Zernike, DoubleZernike
from galsim_test_helpers import *


@timer
def test_Zernike_orthonormality():
    r""" Zernike optical screens *should* be normalized such that
    \int_{unit disc} Z(n1, m1) Z(n2, m2) dA = \pi in unit disc coordinates, or alternatively
    = aperture area if radial coordinate is not normalized (i.e., diam != 2).
    """
    jmax = 30  # Going up to 30 filled Zernikes takes about ~1 sec on my laptop
    diam = 4.0
    R_outer = diam/2

    x = np.linspace(-R_outer, R_outer, 256)
    dx = x[1]-x[0]
    x, y = np.meshgrid(x, x)
    w = np.hypot(x, y) <= R_outer
    x = x[w].ravel()
    y = y[w].ravel()
    area = np.pi*R_outer**2
    for j1 in range(1, jmax+1):
        Z1 = Zernike([0]*(j1+1)+[1], R_outer=R_outer)
        val1 = Z1.evalCartesian(x, y)
        for j2 in range(j1, jmax+1):
            Z2 = Zernike([0]*(j2+1)+[1], R_outer=R_outer)
            val2 = Z2.evalCartesian(x, y)
            integral = np.dot(val1, val2) * dx**2
            if j1 == j2:
                # Only passes at ~1% level because of pixelization.
                np.testing.assert_allclose(
                        integral, area, rtol=1e-2,
                        err_msg="Orthonormality failed for (j1,j2) = ({0},{1})".format(j1, j2))
            else:
                # Only passes at ~1% level because of pixelization.
                np.testing.assert_allclose(
                        integral, 0.0, atol=area*1e-2,
                        err_msg="Orthonormality failed for (j1,j2) = ({0},{1})".format(j1, j2))

    check_pickle(Z1)
    check_pickle(Z1, lambda z: tuple(z.evalCartesian(x, y)))

    # Repeat for Annular Zernikes
    jmax = 22  # Going up to 22 annular Zernikes takes about ~1 sec on my laptop
    R_inner = 0.6
    x = np.linspace(-R_outer, R_outer, 256)
    dx = x[1]-x[0]
    x, y = np.meshgrid(x, x)
    r = np.hypot(x, y)
    w = np.logical_and(R_inner <= r, r <= R_outer)
    x = x[w].ravel()
    y = y[w].ravel()
    area = np.pi*(R_outer**2 - R_inner**2)
    for j1 in range(1, jmax+1):
        Z1 = Zernike([0]*(j1+1)+[1], R_outer=R_outer, R_inner=R_inner)
        val1 = Z1.evalCartesian(x, y)
        for j2 in range(j1, jmax+1):
            Z2 = Zernike([0]*(j2+1)+[1], R_outer=R_outer, R_inner=R_inner)
            val2 = Z2.evalCartesian(x, y)
            integral = np.dot(val1, val2) * dx**2
            if j1 == j2:
                # Only passes at ~1% level because of pixelization.
                np.testing.assert_allclose(
                        integral, area, rtol=1e-2,
                        err_msg="Orthonormality failed for (j1,j2) = ({0},{1})".format(j1, j2))
            else:
                # Only passes at ~1% level because of pixelization.
                np.testing.assert_allclose(
                        integral, 0.0, atol=area*1e-2,
                        err_msg="Orthonormality failed for (j1,j2) = ({0},{1})".format(j1, j2))
    check_pickle(Z1)
    check_pickle(Z1, lambda z: tuple(z.evalCartesian(x, y)))

    with assert_raises(ValueError):
        Z1 = Zernike([0]*4 + [0.1]*7, R_outer=R_inner, R_inner=R_outer)
        val1 = Z1.evalCartesian(x, y)


@timer
def test_annular_Zernike_limit():
    """Check that annular Zernike matches circular Zernike in the limit of 0.0 obscuration.
    """
    jmax = 20
    bd = galsim.BaseDeviate(1029384756)
    u = galsim.UniformDeviate(bd)

    diam = 4.0

    for i in range(4):  # Do a few random tests.  Takes about 1 sec.
        aberrations = [0]+[u() for i in range(jmax)]
        psf1 = galsim.OpticalPSF(diam=diam, lam=500, obscuration=1e-5,
                                 aberrations=aberrations, annular_zernike=True)
        psf2 = galsim.OpticalPSF(diam=diam, lam=500, obscuration=1e-5,
                                 aberrations=aberrations)
        im1 = psf1.drawImage()
        im2 = psf2.drawImage(image=im1.copy())
        # We want the images to be close, since the obscuration is near 0, but not identical.
        # That way we know that the `annular_zernike` keyword is doing something.
        assert im1 != im2, "annular Zernike identical to circular Zernike"
        np.testing.assert_allclose(
                im1.array, im2.array, atol=1e-10,
                err_msg="annular Zernike with 1e-5 obscuration not close to circular Zernike")

    check_pickle(psf1._aper)
    check_pickle(psf1)
    check_pickle(psf1, lambda x: x.drawImage())
    check_pickle(psf2)
    check_pickle(psf2, lambda x: x.drawImage())


@timer
def test_noll():
    """Test that Noll indexing scheme between j <-> (n,m) works as expected.
    """
    # This function stolen from https://github.com/tvwerkhoven/libtim-py/blob/master/libtim/zern.py
    # It used to be in zernike.py, but now we use a faster lookup-table implementation.
    # This reference version is still useful as a test.
    def noll_to_zern(j):
        if (j == 0):
            raise ValueError("Noll indices start at 1. 0 is invalid.")
        n = 0
        j1 = j-1
        while (j1 > n):
            n += 1
            j1 -= n
        m = (-1)**j * ((n % 2) + 2 * int((j1+((n+1) % 2)) / 2.0))
        return (n, m)

    # Test that the version of noll_to_zern in zernike.py is accurate.
    for j in range(1,30):
        true_n,true_m = noll_to_zern(j)
        n,m = galsim.zernike.noll_to_zern(j)
        #print('j=%d, noll = %d,%d, true_noll = %d,%d'%(j,n,m,true_n,true_m))
        assert n == true_n
        assert m == true_m
        # These didn't turn out to be all that useful for fast conversions, but they're cute.
        assert n == int(np.sqrt(8*j-7)-1)//2
        mm = -m if (n//2)%2 == 0 else m
        assert j == n*(n+1)/2 + (abs(2*mm+1)+1)//2

    # Again, the reference version of this function used to be in zernike.py
    def zern_rho_coefs(n, m):
        """Compute coefficients of radial part of Zernike (n, m).
        """
        from galsim.utilities import nCr
        kmax = (n-abs(m)) // 2
        A = np.zeros(n+1)
        for k in range(kmax+1):
            A[n-2*k] = (-1)**k * nCr(n-k, k) * nCr(n-2*k, kmax-k)
        return A

    for j in range(1,30):
        n,m = galsim.zernike.noll_to_zern(j)
        true_coefs = zern_rho_coefs(n,m)
        coefs = galsim.zernike._zern_rho_coefs(n,m)
        #print('j=%d, coefs = %s'%(j,coefs))
        np.testing.assert_array_equal(coefs,true_coefs)


@timer
def test_Zernike_rotate():
    """Test that rotating Zernike coefficients to another coord sys works as expected"""
    #First check that invalid Zernike rotation matrix sizes are trapped
    with assert_raises(ValueError):
        # Can't do size=2, since Z2 mixes into Z3
        galsim.zernike.zernikeRotMatrix(2, 0.1)
        # Can't do size=5, since Z5 mixes into Z6
        galsim.zernike.zernikeRotMatrix(5, 0.2)

    u = galsim.UniformDeviate(12020569031)
    #Now let's test some actual rotations.
    for jmax in [1, 3, 10, 11, 13, 21, 22, 34]:
        # Pick some arbitrary eps and diams
        eps = (jmax % 5)/10.0
        diam = ((jmax % 10) + 1.0)
        # Test points
        rhos = np.linspace(0, diam/2, 4)
        thetas = np.linspace(0, np.pi, 4)

        R_outer = diam/2.0
        R_inner = R_outer*eps

        coefs = [u() for _ in range(jmax+1)]
        Z = Zernike(coefs, R_outer=R_outer, R_inner=R_inner)
        check_pickle(Z)

        for theta in [0.0, 0.1, 1.0, np.pi, 4.0]:
            R = galsim.zernike.zernikeRotMatrix(jmax, theta)
            rotCoefs = np.dot(R, coefs)
            Zrot = Zernike(rotCoefs, R_outer=R_outer, R_inner=R_inner)
            print('j,theta: ',jmax,theta)
            print('Z: ',Z.evalPolar(rhos, thetas))
            print('Zrot: ',Zrot.evalPolar(rhos, thetas+theta))
            print('max diff= ',np.max(np.abs(Zrot.evalPolar(rhos, thetas+theta)-Z.evalPolar(rhos, thetas))))
            np.testing.assert_allclose(
                Z.evalPolar(rhos, thetas),
                Zrot.evalPolar(rhos, thetas+theta),
                atol=1e-12, rtol=0
            )

            Zrot2 = Z.rotate(theta)
            print('Zrot2: ',Zrot2.evalPolar(rhos, thetas+theta))
            print('max diff= ',np.max(np.abs(Zrot2.evalPolar(rhos, thetas+theta)-Z.evalPolar(rhos, thetas))))
            np.testing.assert_allclose(
                Z.evalPolar(rhos, thetas),
                Zrot2.evalPolar(rhos, thetas+theta),
                atol=1e-12, rtol=0
            )


@timer
def test_zernike_eval():
    for coef in [
        np.ones(4),
        np.ones(4, dtype=float),
        np.ones(4, dtype=np.float32)
    ]:
        Z = Zernike(coef)
        assert Z.coef.dtype == np.float64
        assert Z(0.0, 0.0) == 1.0
        assert Z(0, 0) == 1.0

    for coefs in [
        np.ones((4, 4)),
        np.ones((4, 4), dtype=float),
        np.ones((4, 4), dtype=np.float32)
    ]:
        dz = DoubleZernike(coefs)
        assert dz.coef.dtype == np.float64
        assert dz(0.0, 0.0) == dz(0, 0)

        # Make sure we cast to float in _from_uvxy
        uvxy = dz._coef_array_uvxy
        dz2 = DoubleZernike._from_uvxy(uvxy.astype(int))
        np.testing.assert_array_equal(dz2._coef_array_uvxy, dz._coef_array_uvxy)


@timer
def test_ne():
    objs = [
        Zernike([0, 1, 2]),
        Zernike([0, 1, 2, 3]),
        Zernike([0, 1, 2, 3], R_outer=0.2),
        Zernike([0, 1, 2, 3], R_outer=0.2, R_inner=0.1),
        DoubleZernike(np.eye(3)),
        DoubleZernike(np.ones((4, 4))),
        DoubleZernike(np.ones((4, 4)), xy_outer=1.1),
        DoubleZernike(np.ones((4, 4)), xy_outer=1.1, xy_inner=0.9),
        DoubleZernike(np.ones((4, 4)), xy_outer=1.1, xy_inner=0.9, uv_outer=1.1),
        DoubleZernike(np.ones((4, 4)), xy_outer=1.1, xy_inner=0.9, uv_outer=1.1, uv_inner=0.9)
    ]
    check_all_diff(objs)


@timer
def test_Zernike_basis():
    """Test the zernikeBasis function"""
    diam = 2.4
    jmax = 30
    R_outer = diam/2
    R_inner = R_outer*0.2

    u = galsim.UniformDeviate(4669201609)
    for i in range(10):
        # Test at some random points
        x = np.empty((10000,), dtype=float)
        y = np.empty((10000,), dtype=float)
        u.generate(x)
        u.generate(y)

        # zBases will generate all basis vectors at once
        zBases = galsim.zernike.zernikeBasis(jmax, x, y, R_outer=R_outer, R_inner=R_inner)

        # Compare to basis vectors generated one at a time
        for j in range(1, jmax):
            Z = Zernike([0]*j+[1], R_outer=R_outer, R_inner=R_inner)
            zBasis = Z.evalCartesian(x, y)
            np.testing.assert_allclose(
                    zBases[j],
                    zBasis,
                    atol=1e-12, rtol=0)


@timer
def test_fit():
    """Test fitting values to a Zernike series, using the ZernikeBasis function"""
    u = galsim.UniformDeviate(161803)
    for i in range(10):
        x = np.empty((1000,), dtype=float)
        y = np.empty((1000,), dtype=float)
        u.generate(x)
        u.generate(y)
        x -= 0.5
        y -= 0.5
        R_outer = (i%5/5.0)+1
        R_inner = ((i%3/6.0)+0.1)*(R_outer)
        x *= R_outer
        y *= R_outer

        # Should be able to fit quintic polynomial by including Zernikes up to Z_21
        cartesian_coefs = [[u()-0.5, u()-0.5, u()-0.5, u()-0.5, u()-0.5],
                           [u()-0.5, u()-0.5, u()-0.5, u()-0.5,       0],
                           [u()-0.5, u()-0.5, u()-0.5,       0,       0],
                           [u()-0.5, u()-0.5,       0,       0,       0],
                           [u()-0.5,       0,       0,       0,       0]]
        z = galsim.utilities.horner2d(x, y, cartesian_coefs)
        z2 = galsim.utilities.horner2d(x, y, cartesian_coefs, triangle=True)
        np.testing.assert_equal(z,z2)

        basis = galsim.zernike.zernikeBasis(21, x, y, R_outer=R_outer, R_inner=R_inner)
        coefs, _, _, _ = np.linalg.lstsq(basis.T, z, rcond=-1.)
        resids = (Zernike(coefs, R_outer=R_outer, R_inner=R_inner)
                  .evalCartesian(x, y)
                  - z)
        resids2 = np.dot(basis.T, coefs).T - z
        assert resids.shape == x.shape
        assert resids2.shape == x.shape

        np.testing.assert_allclose(resids, 0, atol=1e-14)
        np.testing.assert_allclose(resids2, 0, atol=1e-14)

        # import matplotlib.pyplot as plt
        # fig, axes = plt.subplots(ncols=2, figsize=(8, 4))
        # scat1 = axes[0].scatter(x, y, c=z)
        # plt.colorbar(scat1, ax=axes[0])
        # scat2 = axes[1].scatter(x, y, c=resids)
        # plt.colorbar(scat2, ax=axes[1])
        # plt.show()
        # print(np.mean(resids), np.std(resids))

    # Should also work, and make congruent output, if the shapes of x and y are multi-dimensional
    for i in range(10):
        x = np.empty((1000,), dtype=float)
        y = np.empty((1000,), dtype=float)
        u.generate(x)
        u.generate(y)
        x -= 0.5
        y -= 0.5
        R_outer = (i%5/5.0)+1
        R_inner = ((i%3/6.0)+0.1)*(R_outer)
        x *= R_outer
        y *= R_outer
        x = x.reshape(25, 40)
        y = y.reshape(25, 40)

        # Should be able to fit quintic polynomial by including Zernikes up to Z_21
        cartesian_coefs = [[u()-0.5, u()-0.5, u()-0.5, u()-0.5, u()-0.5],
                           [u()-0.5, u()-0.5, u()-0.5, u()-0.5,       0],
                           [u()-0.5, u()-0.5, u()-0.5,       0,       0],
                           [u()-0.5, u()-0.5,       0,       0,       0],
                           [u()-0.5,       0,       0,       0,       0]]
        z = galsim.utilities.horner2d(x, y, cartesian_coefs)
        assert z.shape == (25, 40)
        z2 = galsim.utilities.horner2d(x, y, cartesian_coefs, triangle=True)
        np.testing.assert_equal(z,z2)

        basis = galsim.zernike.zernikeBasis(21, x, y, R_outer=R_outer, R_inner=R_inner)
        assert basis.shape == (22, 25, 40)
        # lstsq doesn't handle the extra dimension though...
        coefs, _, _, _ = np.linalg.lstsq(basis.reshape(21+1, 1000).T, z.ravel(), rcond=-1.)
        resids = (Zernike(coefs, R_outer=R_outer, R_inner=R_inner)
                  .evalCartesian(x, y)
                  - z)
        resids2 = np.dot(basis.T, coefs).T - z
        assert resids.shape == resids2.shape == x.shape

        np.testing.assert_allclose(resids, 0, atol=1e-14)
        np.testing.assert_allclose(resids2, 0, atol=1e-14)


@timer
def test_gradient():
    """Test that .gradX and .gradY properties work as expected.
    """
    # Start with a few that just quote the literature, e.g., Stephenson (2014).

    Z11 = Zernike([0]*11+[1])

    x = np.linspace(-1, 1, 100)
    x, y = np.meshgrid(x, x)

    def Z11_grad(x, y):
        # Z11 = sqrt(5) (6(x^2+y^2)^2 - 6(x^2+y^2)+1)
        r2 = x**2 + y**2
        gradx = 12*np.sqrt(5)*x*(2*r2-1)
        grady = 12*np.sqrt(5)*y*(2*r2-1)
        return gradx, grady

    # import matplotlib.pyplot as plt
    # fig, axes = plt.subplots(ncols=3, figsize=(12, 3))
    # scat0 = axes[0].scatter(x, y, c=Z11.evalCartesianGrad(x, y)[0])
    # fig.colorbar(scat0, ax=axes[0])
    # scat1 = axes[1].scatter(x, y, c=Z11_grad(x, y)[0])
    # fig.colorbar(scat1, ax=axes[1])
    # scat2 = axes[2].scatter(x, y, c=Z11.evalCartesianGrad(x, y)[0] - Z11_grad(x, y)[0])
    # fig.colorbar(scat2, ax=axes[2])
    # plt.show()

    np.testing.assert_allclose(Z11.evalCartesianGrad(x, y), Z11_grad(x, y), rtol=1.e-12, atol=1e-12)

    Z28 = Zernike([0]*28+[1])

    def Z28_grad(x, y):
        # Z28 = sqrt(14) (x^6 - 15 x^4 y^2 + 15 x^2 y^4 - y^6)
        gradx = 6*np.sqrt(14)*x*(x**4 - 10*x**2*y**2 + 5*y**4)
        grady = -6*np.sqrt(14)*y*(5*x**4 - 10*x**2*y**2 + y**4)
        return gradx, grady

    np.testing.assert_allclose(Z28.evalCartesianGrad(x, y), Z28_grad(x, y), rtol=1.e-12, atol=1e-12)

    # Now try some finite differences on a broader set of input

    def finite_difference_gradient(Z, x, y):
        dh = 1e-5
        return ((Z.evalCartesian(x+dh, y)-Z.evalCartesian(x-dh, y))/(2*dh),
                (Z.evalCartesian(x, y+dh)-Z.evalCartesian(x, y-dh))/(2*dh))

    u = galsim.UniformDeviate(1234)

    # Test finite difference against analytic result for 25 different Zernikes with random number of
    # random coefficients and random inner/outer radii.
    for j in range(25):
        nj = 1+int(u()*55)
        R_inner = 0.2+0.6*u()
        R_outer = R_inner + 0.2+0.6*u()
        Z = Zernike([0]+[u() for _ in range(nj)], R_inner=R_inner, R_outer=R_outer)

        np.testing.assert_allclose(
                finite_difference_gradient(Z, x, y),
                Z.evalCartesianGrad(x, y),
                rtol=1e-5, atol=1e-5)

    # Make sure the gradient of the zero-Zernike works
    Z = Zernike([0])
    assert Z == Z.gradX == Z.gradX.gradX == Z.gradY == Z.gradY.gradY


@timer
def test_gradient_bases():
    """Test the zernikeGradBases function"""
    diam = 2.4
    jmax = 36
    R_outer = diam/2
    R_inner = R_outer*0.2

    u = galsim.UniformDeviate(1029384756)
    for i in range(10):
        # Test at some random points
        x = np.empty((10000,), dtype=float)
        y = np.empty((10000,), dtype=float)
        u.generate(x)
        u.generate(y)

        dxBasis, dyBasis = galsim.zernike.zernikeGradBases(
            jmax, x, y, R_outer=R_outer, R_inner=R_inner
        )

        # Compare to basis vectors generated one at a time
        for j in range(1, jmax+1):
            Z = Zernike([0]*j+[1], R_outer=R_outer, R_inner=R_inner)
            ZX = Z.gradX
            ZY = Z.gradY

            dx = ZX.evalCartesian(x, y)
            dy = ZY.evalCartesian(x, y)

            np.testing.assert_allclose(
                dx, dxBasis[j],
                atol=1e-11, rtol=1e-11
            )

            np.testing.assert_allclose(
                dy, dyBasis[j],
                atol=1e-11, rtol=1e-11
            )


@timer
def test_sum():
    """Test that __add__, __sub__, and __neg__ all work as expected.
    """
    u = galsim.UniformDeviate(5)
    x = np.empty(100, dtype=float)
    y = np.empty(100, dtype=float)
    u.generate(x)
    u.generate(y)

    for _ in range(100):
        n1 = int(u()*53)+3
        n2 = int(u()*53)+3
        R_outer = 1+0.1*u()
        R_inner = 0.1*u()
        if n1 > n2:
            n1, n2 = n2, n1
        a1 = np.empty(n1, dtype=float)
        a2 = np.empty(n2, dtype=float)
        u.generate(a1)
        u.generate(a2)
        z1 = Zernike(a1, R_outer=R_outer, R_inner=R_inner)
        z2 = Zernike(a2, R_outer=R_outer, R_inner=R_inner)

        c1 = u()
        c2 = u()

        coefSum = c2*np.array(z2.coef)
        coefSum[:len(z1.coef)] += c1*z1.coef
        coefDiff = c2*np.array(z2.coef)
        coefDiff[:len(z1.coef)] -= c1*z1.coef
        np.testing.assert_allclose(coefSum, (c1*z1 + c2*z2).coef)
        np.testing.assert_allclose(coefDiff, -(c1*z1 - c2*z2).coef)

        np.testing.assert_allclose(
            c1*z1(x, y) + c2*z2(x, y),
            (c1*z1 + c2*z2)(x, y)
        )
        np.testing.assert_allclose(
            c1*z1(x, y) - c2*z2(x, y),
            (c1*z1 - c2*z2)(x, y)
        )
        # Check that R_outer and R_inner are preserved
        np.testing.assert_allclose(
            (z1+z2).R_outer,
            R_outer
        )
        np.testing.assert_allclose(
            (z1+z2).R_inner,
            R_inner
        )

    with np.testing.assert_raises(TypeError):
        z1 + 3
    with np.testing.assert_raises(TypeError):
        z1 - 3
    with np.testing.assert_raises(ValueError):
        z1 + Zernike([0,1], R_outer=z1.R_outer*2)
    with np.testing.assert_raises(ValueError):
        z1 + Zernike([0,1], R_outer=z1.R_outer, R_inner=z1.R_inner*2)

    # Commutative with integer coefficients
    z1 = Zernike([0,1,2,3,4])
    z2 = Zernike([1,2,3,4,5,6])
    assert z1+z2 == z2+z1
    assert (z2-z1) == z2 + -z1 == -(z1-z2)


@timer
def test_product():
    """Test that __mul__ and __rmul__ work as expected.
    """
    u = galsim.UniformDeviate(57)
    x = np.empty(100, dtype=float)
    y = np.empty(100, dtype=float)
    u.generate(x)
    u.generate(y)

    for _ in range(100):
        n1 = int(u()*21)+3
        n2 = int(u()*21)+3
        R_outer = 1+0.1*u()
        R_inner = 0.1*u()
        a1 = np.empty(n1, dtype=float)
        a2 = np.empty(n2, dtype=float)
        u.generate(a1)
        u.generate(a2)
        z1 = Zernike(a1, R_outer=R_outer, R_inner=R_inner)
        z2 = Zernike(a2, R_outer=R_outer, R_inner=R_inner)

        np.testing.assert_allclose(
            z1(x, y) * z2(x, y),
            (z1 * z2)(x, y),
        )
        np.testing.assert_allclose(
            z1(x, y) * z2(x, y),
            (z2 * z1)(x, y),
        )
        # Check scalar multiplication
        np.testing.assert_allclose(
            (2*z1)(x, y),
            2*(z1(x, y)),
        )
        np.testing.assert_allclose(
            (z1*3.3)(x, y),
            3.3*(z1(x, y)),
        )
        # Check when .coef is missing
        del z1.coef
        np.testing.assert_allclose(
            (z1*3.5)(x, y),
            3.5*(z1(x, y)),
        )
        # Check that R_outer and R_inner are preserved
        np.testing.assert_allclose(
            (z1*z2).R_outer,
            R_outer
        )
        np.testing.assert_allclose(
            (z1*z2).R_inner,
            R_inner
        )
        # Check div
        np.testing.assert_allclose(
            (z1/5.6)(x, y),
            z1(x, y)/5.6,
        )

    with np.testing.assert_raises(TypeError):
        z1 * galsim.Gaussian(fwhm=1)
    with np.testing.assert_raises(ValueError):
        z1 * Zernike([0,1], R_outer=z1.R_outer*2)
    with np.testing.assert_raises(ValueError):
        z1 * Zernike([0,1], R_outer=z1.R_outer, R_inner=z1.R_inner*2)
    with np.testing.assert_raises(TypeError):
        z1 / z2

    # Commutative with integer coefficients
    z1 = Zernike([0,1,2,3,4,5])
    z2 = Zernike([1,2,3,4,5,6])
    assert z1*z2 == z2*z1


@timer
def test_laplacian():
    """Test .laplacian property.
    """
    u = galsim.UniformDeviate(577)
    x = np.empty(100, dtype=float)
    y = np.empty(100, dtype=float)
    u.generate(x)
    u.generate(y)

    for _ in range(200):
        n = int(u()*21)+3
        a = np.empty(n, dtype=float)
        u.generate(a)
        R_outer = 1+0.1*u()
        R_inner = 0.1*u()
        z = Zernike(a, R_outer=R_outer, R_inner=R_inner)

        np.testing.assert_allclose(
            z.laplacian(x, y),
            z.gradX.gradX(x, y) + z.gradY.gradY(x, y)
        )
        # Check that R_outer and R_inner are preserved
        np.testing.assert_allclose(
            z.laplacian.R_outer,
            R_outer
        )
        np.testing.assert_allclose(
            z.laplacian.R_inner,
            R_inner
        )

    # Do a couple by hand
    # Z4 = sqrt(3) (2x^2 + 2y^2 - 1)
    # implies laplacian = 4 sqrt(3) + 4 sqrt(3) = 8 sqrt(3)
    # which is 8 sqrt(3) Z1
    np.testing.assert_allclose(
        Zernike([0,0,0,0,1]).laplacian.coef,
        np.array([0,8*np.sqrt(3)])
    )

    # Z7 = sqrt(8) * (3 * (x^2 + y^2) - 2) * y
    # implies d^2/dx^2 = 6 sqrt(8) y
    #         d^2/dy^2 = 12 sqrt(8) y + 6 sqrt(8) y = 18 sqrt(8) y
    # implies laplacian = 24 sqrt(8) y
    # which is 12*sqrt(8) * Z3 since Z3 = 2 y
    np.testing.assert_allclose(
        Zernike([0,0,0,0,0,0,0,1]).laplacian.coef,
        np.array([0,0,0,12*np.sqrt(8)])
    )


@timer
def test_hessian():
    """Test .hessian property.
    """
    u = galsim.UniformDeviate(5772)
    x = np.empty(100, dtype=float)
    y = np.empty(100, dtype=float)
    u.generate(x)
    u.generate(y)

    for _ in range(200):
        n = int(u()*21)+3
        a = np.empty(n, dtype=float)
        u.generate(a)
        R_outer = 1+0.1*u()
        R_inner = 0.1*u()
        z = Zernike(a, R_outer=R_outer, R_inner=R_inner)

        np.testing.assert_allclose(
            z.hessian(x, y),
            z.gradX.gradX(x, y) * z.gradY.gradY(x, y) - z.gradX.gradY(x, y)**2
        )
        # Check that R_outer and R_inner are preserved
        np.testing.assert_allclose(
            z.hessian.R_outer,
            R_outer
        )
        np.testing.assert_allclose(
            z.hessian.R_inner,
            R_inner
        )

    # Do a couple by hand
    # Z4 = sqrt(3) (2x^2 + 2y^2 - 1)
    # implies hessian = 4 sqrt(3) * 4 sqrt(3) - 0 * 0 = 16*3 = 48
    # which is 48 Z1
    np.testing.assert_allclose(
        Zernike([0,0,0,0,1]).hessian.coef,
        np.array([0,48])
    )

    # Z7 = sqrt(8) * (3 * (x^2 + y^2) - 2) * y
    # implies d^2/dx^2 = 6 sqrt(8) y
    #         d^2/dy^2 = 12 sqrt(8) y + 6 sqrt(8) y = 18 sqrt(8) y
    #         d^2/dxdy = 6 sqrt(8) x
    # implies hessian = 6 sqrt(8) y * 18 sqrt(8) y - (6 sqrt(8) x)^2
    #                 = 108 * 8 * y^2 - 36 * 8 x^2 = 864 y^2 - 288 x^2
    # That's a little inconvenient to decompose into Zernikes by hand, but we can test against
    # an array of (x,y) values.
    np.testing.assert_allclose(
        Zernike([0,0,0,0,0,0,0,1]).hessian(x, y),
        864*y*y - 288*x*x
    )


@timer
def test_describe_zernike():
    """Test that Zernike descriptions make sense."""
    # Just do a few by hand
    # These can be looked up in Lakshminarayanan & Fleck (2011), Journal of Modern Optics
    # Table 1 there has algebraic expressions for Zernikes through j=36
    # Note, their definition is slightly different than ours: x and y are swapped.  (See their
    # figure 2 in which the azimuthal angle is defined +ve CW from +y.  We use +ve CCW from +x to be
    # consistent with Zemax.)
    assert galsim.zernike.describe_zernike(1) == "sqrt(1) * (1)"
    assert galsim.zernike.describe_zernike(2) == "sqrt(4) * (x)"
    assert galsim.zernike.describe_zernike(3) == "sqrt(4) * (y)"
    assert galsim.zernike.describe_zernike(4) == "sqrt(3) * (-1 + 2y^2 + 2x^2)"
    assert galsim.zernike.describe_zernike(10) == "sqrt(8) * (-3xy^2 + x^3)"

    Z22str = (
        "sqrt(7) * (-1 + 12y^2 - 30y^4 + 20y^6 + 12x^2"
        " - 60x^2y^2 + 60x^2y^4 - 30x^4 + 60x^4y^2 + 20x^6)"
    )
    assert galsim.zernike.describe_zernike(22) == Z22str

    Z36str = (
        "sqrt(16) * (-7xy^6 + 35x^3y^4 - 21x^5y^2 + x^7)"
    )
    assert galsim.zernike.describe_zernike(36) == Z36str


@timer
def test_lazy_coef():
    """Check that coefs reconstructed from _coef_array_xy round trip correctly."""
    bd = galsim.BaseDeviate(191120)
    u = galsim.UniformDeviate(bd)
    # For triangular jmax, get the same shape array back.
    for jmax in [3, 6, 10, 15, 21]:
        zarr = [0]+[u() for i in range(jmax)]
        R_inner = u()*0.5+0.2
        R_outer = u()*2.0+2.0
        Z = Zernike(zarr, R_outer=R_outer, R_inner=R_inner)
        Z._coef_array_xy
        del Z.coef
        np.testing.assert_allclose(zarr, Z.coef, rtol=0, atol=1e-12)

    # For non-triangular jmax, get shape rounded up to next triangular
    for jmax in [2, 7, 11, 17, 23]:
        zarr = [0]+[u() for i in range(jmax)]
        R_inner = u()*0.5+0.2
        R_outer = u()*2.0+2.0
        Z = Zernike(zarr, R_outer=R_outer, R_inner=R_inner)
        Z._coef_array_xy
        del Z.coef
        np.testing.assert_allclose(zarr, Z.coef[:len(zarr)], rtol=0, atol=1e-12)
        # extra coefficients are all ~0
        np.testing.assert_allclose(Z.coef[len(zarr):], 0.0, rtol=0, atol=1e-12)


@timer
def test_dz_val():
    rng = galsim.BaseDeviate(1234).as_numpy_generator()
    for _ in range(10):
        kmax = rng.integers(4, 12)
        jmax = rng.integers(4, 12)
        coef = rng.normal(size=(kmax+1, jmax+1))
        uv_inner = rng.uniform(0.4, 0.7)
        uv_outer = rng.uniform(1.3, 1.7)
        xy_inner = rng.uniform(0.4, 0.7)
        xy_outer = rng.uniform(1.3, 1.7)
        dz = DoubleZernike(
            coef,
            uv_inner=uv_inner,
            uv_outer=uv_outer,
            xy_inner=xy_inner,
            xy_outer=xy_outer,
        )

        uv_scalar = rng.normal(size=(2,))
        xy_scalar = rng.normal(size=(2,))
        uv_vector = rng.normal(size=(2, 10))
        xy_vector = rng.normal(size=(2, 10))

        check_pickle(dz)
        check_pickle(dz, lambda dz_: dz_.coef.shape)
        check_pickle(dz, lambda dz_: tuple(dz_.coef.ravel()))
        check_pickle(dz, lambda dz_: dz_._coef_array_uvxy.shape)
        check_pickle(dz, lambda dz_: tuple(dz_._coef_array_uvxy.ravel()))
        check_pickle(dz, lambda dz_: dz_(*uv_scalar))
        check_pickle(dz, lambda dz_: tuple(dz_(*uv_vector)))
        check_pickle(dz, lambda dz_: dz_(*uv_scalar, *xy_scalar))
        check_pickle(dz, lambda dz_: tuple(dz_(*uv_vector, *xy_vector)))

        # If you don't specify xy, then get (list of) Zernike out.
        assert isinstance(dz(*uv_scalar), Zernike)
        assert isinstance(dz(*uv_vector), list)
        assert all(isinstance(z, Zernike) for z in dz(*uv_vector))

        # If uv scalar and xy scalar, then get scalar out.
        assert np.ndim(dz(*uv_scalar, *xy_scalar)) == 0
        # If uv vector and xy scalar, then get vector out.
        assert np.ndim(dz(*uv_vector, *xy_scalar)) == 1
        # If uv scalar and xy vector, then get vector out.
        assert np.ndim(dz(*uv_scalar, *xy_vector)) == 1
        # If uv vector and xy vector, then get vector out.
        assert np.ndim(dz(*uv_vector, *xy_vector)) == 1

        # Check consistency of __call__ outputs
        zk_list = dz(*uv_vector)
        vals = dz(*uv_vector, *xy_vector)
        np.testing.assert_allclose(
            np.array([zk(x, y) for x, y, zk in zip(*xy_vector, zk_list)]),
            vals,
            atol=2e-13, rtol=0
        )
        for i, (x, y) in enumerate(xy_vector.T):
            np.testing.assert_allclose(
                vals[i],
                zk_list[i](x, y),
                atol=2e-13, rtol=0
            )
        for i, (u, v, x, y) in enumerate(zip(*uv_vector, *xy_vector)):
            np.testing.assert_allclose(
                vals[i],
                dz(u, v, x, y),
                atol=2e-13, rtol=0
            )
        for i, (u, v) in enumerate(zip(*uv_vector)):
            np.testing.assert_allclose(
                vals[i],
                dz(u, v)(*xy_vector[:, i]),
                atol=2e-13, rtol=0
            )

        zk_coefs = dz.xycoef(*uv_vector)
        for zk, c in zip(zk_list, zk_coefs):
            # Zk may have trailing zeros...
            ncoef = len(c)
            np.testing.assert_allclose(
                zk.coef[:ncoef],
                c,
                atol=2e-13, rtol=0
            )
        for i, (u, v) in enumerate(zip(*uv_vector)):
            np.testing.assert_equal(
                zk_coefs[i],
                dz.xycoef(u, v)
            )

        # Check asserts
        with assert_raises(AssertionError):
            dz(0.0, [1.0])
        with assert_raises(AssertionError):
            dz([0.0], [1.0, 1.0])
        with assert_raises(galsim.GalSimIncompatibleValuesError):
            dz(0.0, 0.0, x=0.0, y=None)
        with assert_raises(AssertionError):
            dz(0.0, 0.0, x=[1.0], y=1.0)
        with assert_raises(AssertionError):
            dz(0.0, 0.0, x=[1.0], y=[1.0, 2.0])
        with assert_raises(AssertionError):
            dz([0.0, 1.0], [0.0, 1.0], x=[1.0], y=[1.0])

    # Try pickle/repr with default domain
    dz = DoubleZernike(coef)
    check_pickle(dz)


@timer
def test_dz_coef_uvxy():
    rng = galsim.BaseDeviate(4321).as_numpy_generator()
    for _ in range(100):
        kmax = rng.integers(4, 22)
        jmax = rng.integers(4, 22)
        coef = rng.normal(size=(kmax+1, jmax+1))
        coef[0] = 0.0
        coef[:, 0] = 0.0
        uv_inner = rng.uniform(0.4, 0.7)
        uv_outer = rng.uniform(1.3, 1.7)
        xy_inner = rng.uniform(0.4, 0.7)
        xy_outer = rng.uniform(1.3, 1.7)
        dz = DoubleZernike(
            coef,
            uv_inner=uv_inner,
            uv_outer=uv_outer,
            xy_inner=xy_inner,
            xy_outer=xy_outer
        )
        # Test that we can recover coef from coef_array_xyuv
        dz._coef_array_uvxy
        del dz.coef
        np.testing.assert_allclose(
            dz.coef[:coef.shape[0], :coef.shape[1]],
            coef,
            rtol=0,
            atol=1e-12
        )

        uv_scalar = rng.normal(size=(2,))
        xy_scalar = rng.normal(size=(2,))
        uv_vector = rng.normal(size=(2, 10))
        xy_vector = rng.normal(size=(2, 10))

        # Scalar uv only
        zk1 = dz._call_old(*uv_scalar)
        zk2 = dz(*uv_scalar)
        n = len(zk1.coef)
        np.testing.assert_allclose(
            zk1.coef[1:n],
            zk2.coef[1:n],
            rtol=1e-11, atol=1e-11
        )

        # Vector uv only
        zks1 = dz._call_old(*uv_vector)
        zks2 = dz(*uv_vector)
        for zk1, zk2 in zip(zks1, zks2):
            n = len(zk1.coef)
            np.testing.assert_allclose(
                zk1.coef[1:n],
                zk2.coef[1:n],
                rtol=1e-11, atol=1e-11
            )

        # All scalar/vector combinations
        for uv in [uv_scalar, uv_vector]:
            for xy in [xy_scalar, xy_vector]:
                np.testing.assert_allclose(
                    dz(*uv, *xy),
                    dz._call_old(*uv, *xy)
                )


@timer
def test_dz_sum():
    """Test that DZ.__add__, __sub__, and __neg__ work as expected.
    """
    rng = galsim.BaseDeviate(57721).as_numpy_generator()
    u = rng.uniform(-1.0, 1.0, size=100)
    v = rng.uniform(-1.0, 1.0, size=100)
    x = rng.uniform(-1.0, 1.0, size=100)
    y = rng.uniform(-1.0, 1.0, size=100)

    for _ in range(100):
        k1 = rng.integers(1, 11)
        j1 = rng.integers(1, 11)
        k2 = rng.integers(1, 11)
        j2 = rng.integers(1, 11)

        uv_inner = rng.uniform(0.4, 0.7)
        uv_outer = rng.uniform(1.3, 1.7)
        xy_inner = rng.uniform(0.4, 0.7)
        xy_outer = rng.uniform(1.3, 1.7)

        coef1 = rng.normal(size=(k1+1, j1+1))
        coef1[0] = 0.0
        coef1[:, 0] = 0.0
        coef2 = rng.normal(size=(k2+1, j2+1))
        coef2[0] = 0.0
        coef2[:, 0] = 0.0

        dz1 = DoubleZernike(
            coef1,
            uv_inner=uv_inner, uv_outer=uv_outer,
            xy_inner=xy_inner, xy_outer=xy_outer
        )
        dz2 = DoubleZernike(
            coef2,
            uv_inner=uv_inner, uv_outer=uv_outer,
            xy_inner=xy_inner, xy_outer=xy_outer
        )
        c1 = rng.uniform(-1.0, 1.0)
        c2 = rng.uniform(-1.0, 1.0)

        kmax = max(k1, k2)
        jmax = max(j1, j2)

        coefSum = np.zeros((kmax+1, jmax+1))
        coefSum[:k1+1, :j1+1] = c1*coef1
        coefSum[:k2+1, :j2+1] += c2*coef2

        coefDiff = np.zeros((kmax+1, jmax+1))
        coefDiff[:k1+1, :j1+1] = c1*coef1
        coefDiff[:k2+1, :j2+1] -= c2*coef2

        np.testing.assert_allclose(coefSum, (c1*dz1 + c2*dz2).coef)
        np.testing.assert_allclose(coefDiff, (c1*dz1 - c2*dz2).coef)

        np.testing.assert_allclose(
            c1*dz1(u, v, x, y) + c2*dz2(u, v, x, y),
            (c1*dz1 + c2*dz2)(u, v, x, y)
        )
        np.testing.assert_allclose(
            c1*dz1(u, v, x, y) - c2*dz2(u, v, x, y),
            (c1*dz1 - c2*dz2)(u, v, x, y)
        )

        # Check that domains are preserved
        dzsum = dz1 + dz2
        np.testing.assert_allclose(
            dzsum.uv_inner,
            uv_inner
        )
        np.testing.assert_allclose(
            dzsum.uv_outer,
            uv_outer
        )
        np.testing.assert_allclose(
            dzsum.xy_inner,
            xy_inner
        )
        np.testing.assert_allclose(
            dzsum.xy_outer,
            xy_outer
        )

        with np.testing.assert_raises(TypeError):
            dz1 + 1
        with np.testing.assert_raises(TypeError):
            dz1 - 3
        with np.testing.assert_raises(ValueError):
            dz1 + DoubleZernike(
                coef1, uv_outer=2*uv_outer, uv_inner=uv_inner, xy_outer=xy_outer, xy_inner=xy_inner
            )
        with np.testing.assert_raises(ValueError):
            dz1 + DoubleZernike(
                coef1, uv_outer=uv_outer, uv_inner=2*uv_inner, xy_outer=xy_outer, xy_inner=xy_inner
            )
        with np.testing.assert_raises(ValueError):
            dz1 + DoubleZernike(
                coef1, uv_outer=uv_outer, uv_inner=uv_inner, xy_outer=2*xy_outer, xy_inner=xy_inner
            )
        with np.testing.assert_raises(ValueError):
            dz1 + DoubleZernike(
                coef1, uv_outer=uv_outer, uv_inner=uv_inner, xy_outer=xy_outer, xy_inner=2*xy_inner
            )

        # Commutative with integer coefficients
        dz1 = DoubleZernike(np.eye(3, dtype=int))
        dz2 = DoubleZernike(np.ones((4, 4), dtype=int))
        assert dz1 + dz2 == dz2 + dz1
        assert (dz2 - dz1) == dz2 + (-dz1) == -(dz1 - dz2)

        # Check again with missing .coef
        del dz1.coef
        del dz2.coef
        assert dz1 + dz2 == dz2 + dz1
        assert (dz2 - dz1) == dz2 + (-dz1) == -(dz1 - dz2)


@timer
def test_dz_product():
    """Test that __mul__ and __rmul__ work as expected.
    """
    rng = galsim.BaseDeviate(31415).as_numpy_generator()
    u = rng.uniform(-1.0, 1.0, size=100)
    v = rng.uniform(-1.0, 1.0, size=100)
    x = rng.uniform(-1.0, 1.0, size=100)
    y = rng.uniform(-1.0, 1.0, size=100)

    for _ in range(100):
        k1 = rng.integers(1, 16)
        j1 = rng.integers(1, 16)
        k2 = rng.integers(1, 16)
        j2 = rng.integers(1, 16)

        uv_inner = rng.uniform(0.4, 0.7)
        uv_outer = rng.uniform(1.3, 1.7)
        xy_inner = rng.uniform(0.4, 0.7)
        xy_outer = rng.uniform(1.3, 1.7)

        coef1 = rng.normal(size=(k1+1, j1+1))
        coef1[0] = 0.0
        coef1[:, 0] = 0.0
        coef2 = rng.normal(size=(k2+1, j2+1))
        coef2[0] = 0.0
        coef2[:, 0] = 0.0

        dz1 = DoubleZernike(
            coef1,
            uv_inner=uv_inner, uv_outer=uv_outer,
            xy_inner=xy_inner, xy_outer=xy_outer
        )
        dz2 = DoubleZernike(
            coef2,
            uv_inner=uv_inner, uv_outer=uv_outer,
            xy_inner=xy_inner, xy_outer=xy_outer
        )

        np.testing.assert_allclose(
            dz1(u, v, x, y)*dz2(u, v, x, y),
            (dz1 * dz2)(u, v, x, y)
        )
        np.testing.assert_allclose(
            dz1(u, v, x, y)*dz2(u, v, x, y),
            (dz2 * dz1)(u, v, x, y)
        )
        # Check scalar multiplication
        np.testing.assert_allclose(
            2*dz1(u, v, x, y),
            (2 * dz1)(u, v, x, y)
        )
        np.testing.assert_allclose(
            dz1(u, v, x, y)*3.3,
            (dz1 * 3.3)(u, v, x, y)
        )
        # Try when .coef is missing
        del dz1.coef
        np.testing.assert_allclose(
            dz1(u, v, x, y)*3.5,
            (dz1 * 3.5)(u, v, x, y)
        )
        # Check that domain is preserved
        dzprod = dz1 * dz2
        np.testing.assert_equal(
            [dzprod.uv_inner, dzprod.uv_outer, dzprod.xy_inner, dzprod.xy_outer],
            [uv_inner, uv_outer, xy_inner, xy_outer]
        )
        # Check div
        np.testing.assert_allclose(
            (dz1 / 5.6)(u, v, x, y),
            dz1(u, v, x, y)/5.6,
        )


    with np.testing.assert_raises(TypeError):
        dz1 * galsim.Gaussian(sigma=1.0)
    with np.testing.assert_raises(ValueError):
        dz1 * DoubleZernike(
            coef1, uv_outer=2*uv_outer, uv_inner=uv_inner, xy_outer=xy_outer, xy_inner=xy_inner
        )
    with np.testing.assert_raises(ValueError):
        dz1 * DoubleZernike(
            coef1, uv_outer=uv_outer, uv_inner=2*uv_inner, xy_outer=xy_outer, xy_inner=xy_inner
        )
    with np.testing.assert_raises(ValueError):
        dz1 * DoubleZernike(
            coef1, uv_outer=uv_outer, uv_inner=uv_inner, xy_outer=2*xy_outer, xy_inner=xy_inner
        )
    with np.testing.assert_raises(ValueError):
        dz1 * DoubleZernike(
            coef1, uv_outer=uv_outer, uv_inner=uv_inner, xy_outer=xy_outer, xy_inner=2*xy_inner
        )
    with np.testing.assert_raises(TypeError):
        dz1 / dz2

    # Commutative with integer coefficients
    dz1 = DoubleZernike(np.eye(3, dtype=int))
    dz2 = DoubleZernike(np.ones((4, 4), dtype=int))
    assert dz1 * dz2 == dz2 * dz1
    assert (dz2 * 3) == (3 * dz2)


@timer
def test_dz_grad():
    """Test that DZ gradients work as expected.
    """
    rng = galsim.BaseDeviate(31415).as_numpy_generator()
    u = rng.uniform(-1.0, 1.0, size=100)
    v = rng.uniform(-1.0, 1.0, size=100)
    x = rng.uniform(-1.0, 1.0, size=100)
    y = rng.uniform(-1.0, 1.0, size=100)

    for _ in range(10):
        k1 = rng.integers(1, 16)
        j1 = rng.integers(1, 16)

        uv_inner = rng.uniform(0.4, 0.7)
        uv_outer = rng.uniform(1.3, 1.7)
        xy_inner = rng.uniform(0.4, 0.7)
        xy_outer = rng.uniform(1.3, 1.7)

        coef = rng.normal(size=(k1+1, j1+1))
        coef[0] = 0.0
        coef[:, 0] = 0.0

        dz = DoubleZernike(
            coef,
            uv_inner=uv_inner, uv_outer=uv_outer,
            xy_inner=xy_inner, xy_outer=xy_outer
        )

        # X and Y are easy to check with single Zernike gradient functions.
        np.testing.assert_allclose(
            dz.gradX(u, v, x, y),
            np.array([dz(u_, v_).gradX(x_, y_) for u_, v_, x_, y_ in zip(u, v, x, y)])
        )
        np.testing.assert_allclose(
            dz.gradY(u, v, x, y),
            np.array([dz(u_, v_).gradY(x_, y_) for u_, v_, x_, y_ in zip(u, v, x, y)])
        )
        # U and V are trickier, since we aren't including a way to turn a DZ evaluated
        # at (x, y) into a single Zernike of (u, v).  We can mock that though up by
        # transposing the DZ coefficients and swapping the domain parameters.
        dz_xyuv = DoubleZernike(
            np.transpose(coef, axes=(1, 0)),
            uv_inner=xy_inner, uv_outer=xy_outer,
            xy_inner=uv_inner, xy_outer=uv_outer
        )
        np.testing.assert_allclose(
            dz.gradU(u, v, x, y),
            np.array([dz_xyuv(x_, y_).gradX(u_, v_) for u_, v_, x_, y_ in zip(u, v, x, y)])
        )
        np.testing.assert_allclose(
            dz.gradV(u, v, x, y),
            np.array([dz_xyuv(x_, y_).gradY(u_, v_) for u_, v_, x_, y_ in zip(u, v, x, y)])
        )

        # Test Zernike coefficients themselves.
        for u_, v_ in zip(u, v):
            dzdx1 = dz(u_, v_).gradX.coef
            dzdx2 = dz.gradX(u_, v_).coef
            dzdx1 = np.concatenate([dzdx1, np.zeros(len(dzdx2) - len(dzdx1))])
            np.testing.assert_allclose(dzdx1, dzdx2, atol=1e-12)

            dzdy1 = dz(u_, v_).gradY.coef
            dzdy2 = dz.gradY(u_, v_).coef
            dzdy1 = np.concatenate([dzdy1, np.zeros(len(dzdy2) - len(dzdy1))])
            np.testing.assert_allclose(dzdy1, dzdy2, atol=1e-12)


@timer
def test_dz_to_T():
    """Test that DZs enable efficient computation of optical PSF sizes."""

    # One of the reasons to build the DZ class was to see if we can quickly
    # convert from coefficients that, say, come out of batoid or Piff or an
    # active optics framework into a PSF size, both averaged and over the field
    # of view.

    # The idea is that x and y deflections of photons from their optimal paths
    # are directly proportional to the wavefront gradient.  I.e.,
    # delta_xfp \propto dW(u, v, x, y)/dx
    # where delta_xfp is the focal plane deflection, and W is the double Zernike
    # wavefront which is a function of field angle (u, v) and pupil position
    # (x, y).

    # We want, as a function of (u, v), the variance of delta_xfp.  This is
    # Var[delta_xfp] \propto Var[dW/dx] = E[(dW/dx)^2] - (E[dW/dx])^2,  where E
    # is the expectation value over the pupil.

    # First construct a DZ.
    rng = galsim.BaseDeviate(51413).as_numpy_generator()

    for _ in range(10):
        k1 = rng.integers(1, 29)
        j1 = rng.integers(4, 11)

        uv_inner = rng.uniform(0.4, 0.7)
        uv_outer = rng.uniform(1.3, 1.7)
        xy_inner = rng.uniform(0.4, 0.7)
        xy_outer = rng.uniform(1.3, 1.7)

        coef = rng.normal(size=(k1+1, j1+1))
        coef[0] = 0.0
        coef[:, 0] = 0.0

        W = DoubleZernike(
            coef,
            uv_inner=uv_inner, uv_outer=uv_outer,  # field
            xy_inner=xy_inner, xy_outer=xy_outer   # pupil
        )

        # Now the analytic map of optical PSF size T
        dWdx = W.gradX
        dWdy = W.gradY
        dWdx2 = dWdx * dWdx
        dWdy2 = dWdy * dWdy

        # These are still double Zernikes.  Extract their expectation values
        # over the pupil (which are functions of field angle).
        dWdx_field = dWdx.mean_xy
        dWdy_field = dWdy.mean_xy
        dWdx2_field = dWdx2.mean_xy
        dWdy2_field = dWdy2.mean_xy
        # Now construct the PSF size T
        T = dWdx2_field + dWdy2_field - dWdx_field*dWdx_field - dWdy_field*dWdy_field

        # Now compare to a quick Monte Carlo estimate of T.
        us = []
        vs = []
        Ts = []
        for _ in range(100):  # 100 positions
            # Pick a random field angle:
            uv = np.inf
            while uv > uv_outer or uv < uv_inner:
                u_ = rng.uniform(-uv_outer, uv_outer)
                v_ = rng.uniform(-uv_outer, uv_outer)
                uv = np.hypot(u_, v_)
            us.append(u_)
            vs.append(v_)
            # Get the wavefront at that field angle
            W_ = W(u_, v_)
            # Pick random pupil positions
            dxs = []
            dys = []
            for _ in range(50):  # 50 photons
                xy = np.inf
                while xy > xy_outer or xy < xy_inner:
                    x_ = rng.uniform(-xy_outer, xy_outer)
                    y_ = rng.uniform(-xy_outer, xy_outer)
                    xy = np.hypot(x_, y_)
                # Get the focal plane deflection
                dxs.append(W_.gradX(x_, y_))
                dys.append(W_.gradY(x_, y_))
            # Compute the variance of the focal plane deflection
            Ts.append(np.var(dxs) + np.var(dys))

        # Assert that the median relative error is less than 5%,
        # and the 90% worst case is less than 35%
        np.testing.assert_array_less(
            np.abs(np.quantile((T(us, vs) - Ts)/Ts, [0.5, 0.9])),
            [0.05, 0.35]
        )

        # Uncomment below to look at the plots

        # def colorbar(mappable):
        #     from mpl_toolkits.axes_grid1 import make_axes_locatable
        #     import matplotlib.pyplot as plt
        #     last_axes = plt.gca()
        #     ax = mappable.axes
        #     fig = ax.figure
        #     divider = make_axes_locatable(ax)
        #     cax = divider.append_axes("right", size="5%", pad=0.05)
        #     cbar = fig.colorbar(mappable, cax=cax)
        #     plt.sca(last_axes)
        #     return cbar

        # ugrid = np.linspace(-uv_outer, uv_outer, 100)
        # ugrid, vgrid = np.meshgrid(ugrid, ugrid)
        # Tgrid = T(ugrid, vgrid)
        # w = np.hypot(ugrid, vgrid) > uv_outer
        # w |= np.hypot(ugrid, vgrid) < uv_inner
        # Tgrid[w] = np.nan

        # import matplotlib.pyplot as plt
        # fig, axes = plt.subplots(ncols=2, nrows=1, figsize=(10, 5))
        # vmin, vmax = np.nanquantile(Tgrid, [0.02, 0.98])
        # colorbar(axes[0].scatter(us, vs, c=Ts, vmin=vmin, vmax=vmax))
        # colorbar(axes[1].imshow(Tgrid, extent=(-uv_outer, uv_outer, -uv_outer, uv_outer), origin='lower', vmin=vmin, vmax=vmax))
        # axes[1].scatter(us, vs, c=Ts, vmin=vmin, vmax=vmax)
        # axes[0].set_aspect('equal')
        # fig.tight_layout()
        # plt.show()


@timer
def test_dz_rotate():
    rng = galsim.BaseDeviate(12775).as_numpy_generator()

    for _ in range(30):
        k1 = rng.choice([1, 3, 10, 11, 13, 21, 22, 34])
        j1 = rng.choice([1, 3, 10, 11, 13, 21, 22, 34])

        uv_inner = rng.uniform(0.4, 0.7)
        uv_outer = rng.uniform(1.3, 1.7)
        xy_inner = rng.uniform(0.4, 0.7)
        xy_outer = rng.uniform(1.3, 1.7)

        coef = rng.normal(size=(k1+1, j1+1))
        coef[0] = 0.0
        coef[:, 0] = 0.0

        dz = DoubleZernike(
            coef,
            uv_inner=uv_inner, uv_outer=uv_outer,  # field
            xy_inner=xy_inner, xy_outer=xy_outer   # pupil
        )

        # Test points
        r = rng.uniform(0.0, 1.0, size=30)
        th = rng.uniform(0.0, 2*np.pi, size=30)
        rho = rng.uniform(0.0, 1.0, size=30)
        ph = rng.uniform(0.0, 2*np.pi, size=30)

        for theta_uv in [0.0, 0.2, 0.4]:
            for theta_xy in [0.0, 0.1, 0.3]:
                dz_rot = dz.rotate(theta_uv=theta_uv, theta_xy=theta_xy)
                np.testing.assert_allclose(
                    dz(
                        rho*np.cos(ph), rho*np.sin(ph),
                        r*np.cos(th), r*np.sin(th)
                    ),
                    dz_rot(
                        rho*np.cos(ph+theta_uv), rho*np.sin(ph+theta_uv),
                        r*np.cos(th+theta_xy), r*np.sin(th+theta_xy)
                    ),
                    atol=1e-11, rtol=0
                )


@timer
def test_dz_basis():
    """Test the doubleZernikeBasis function"""

    rng = galsim.BaseDeviate(127750).as_numpy_generator()

    for _ in range(10):
        k1 = rng.choice([3, 7, 10, 15])
        j1 = rng.choice([3, 5, 7])

        uv_inner = rng.uniform(0.4, 0.7)
        uv_outer = rng.uniform(1.3, 1.7)
        xy_inner = rng.uniform(0.4, 0.7)
        xy_outer = rng.uniform(1.3, 1.7)

        for _ in range(10):
            # Test at some random points
            x, y, u, v = rng.uniform(-0.5, 0.5, size=(4, 1000))

            # dzBases will generate all basis vectors at once
            dzBases = galsim.zernike.doubleZernikeBasis(
                k1, j1, x, y, u, v,
                uv_inner=uv_inner, uv_outer=uv_outer,
                xy_inner=xy_inner, xy_outer=xy_outer,
            )

            for j in range(1, j1):
                for k in range(1, k1):
                    coef = np.zeros((k1, j1))
                    coef[k, j] = 1.0
                    DZ = DoubleZernike(
                        coef,
                        uv_inner=uv_inner, uv_outer=uv_outer,
                        xy_inner=xy_inner, xy_outer=xy_outer
                    )
                    DZbasis = DZ(x, y, u, v)
                    np.testing.assert_allclose(
                        DZbasis, dzBases[k, j],
                        atol=1e-11, rtol=0
                    )


@timer
def test_dz_mean():
    """Test the dz.mean_xy and .mean_uv properties"""
    rng = galsim.BaseDeviate(51413).as_numpy_generator()

    for _ in range(10):
        k1 = rng.choice([3, 7, 10, 15])
        j1 = rng.choice([3, 5, 7])

        uv_inner = rng.uniform(0.4, 0.7)
        uv_outer = rng.uniform(1.3, 1.7)
        xy_inner = rng.uniform(0.4, 0.7)
        xy_outer = rng.uniform(1.3, 1.7)

        coef = rng.normal(size=(k1+1, j1+1))
        coef[0] = 0.0
        coef[:, 0] = 0.0

        dz = DoubleZernike(
            coef,
            uv_inner=uv_inner, uv_outer=uv_outer,
            xy_inner=xy_inner, xy_outer=xy_outer
        )
        # We don't have a function that returns a Zernike over uv at a given xy
        # point, but we can mimic that by transposing xy an uv in a new
        # DoubleZernike object.
        dzT = DoubleZernike(
            coef.T,
            uv_inner=xy_inner, uv_outer=xy_outer,
            xy_inner=uv_inner, xy_outer=uv_outer
        )

        mean_xy = dz.mean_xy
        mean_uv = dz.mean_uv

        for _ in range(10):
            # Pick a random uv point
            uv = np.inf
            while uv > uv_outer or uv < uv_inner:
                u = rng.uniform(-uv_outer, uv_outer)
                v = rng.uniform(-uv_outer, uv_outer)
                uv = np.hypot(u, v)
            # Evaluate at that point
            zk = dz(u, v)
            np.testing.assert_allclose(
                zk.coef[1],
                mean_xy(u, v),
            )

        for _ in range(10):
            # Pick a random xy point
            xy = np.inf
            while xy > xy_outer or xy < xy_inner:
                x = rng.uniform(-xy_outer, xy_outer)
                y = rng.uniform(-xy_outer, xy_outer)
                xy = np.hypot(x, y)
            # Evaluate at that point
            zk = dzT(x, y)
            np.testing.assert_allclose(
                zk.coef[1],
                mean_uv(x, y),
            )

        # Make sure we can construct from ._coef_array_uvxy when .coef is
        # unavailable too.
        dz._coef_array_uvxy  # Ensure _coef_array_uvxy exists
        del dz.coef, dz.mean_uv, dz.mean_xy  # Clear out lazy_properties
        mean_xy2 = dz.mean_xy
        mean_uv2 = dz.mean_uv

        sh = min(len(mean_xy.coef), len(mean_xy2.coef))
        np.testing.assert_allclose(
            mean_xy.coef[:sh], mean_xy2.coef[:sh],
            atol=1e-14, rtol=0
        )
        np.testing.assert_equal(
            (mean_xy.R_inner, mean_xy.R_outer),
            (mean_xy2.R_inner, mean_xy2.R_outer),
        )


        sh = min(len(mean_uv.coef), len(mean_uv2.coef))
        np.testing.assert_allclose(
            mean_uv.coef[:sh], mean_uv2.coef[:sh],
            atol=1e-14, rtol=0
        )
        np.testing.assert_equal(
            (mean_uv.R_inner, mean_uv.R_outer),
            (mean_uv2.R_inner, mean_uv2.R_outer),
        )


def test_large_j(run_slow):
    # The analytic form for an annular Zernike of the form (n, m) = (n, n) or (n, -n)
    # is:
    #   r^n sincos(n theta) sqrt((2n+2) / sum_i=0^n-1 eps^(2i))
    # where sincos is sin if n is even and cos if n is odd, and eps = R_inner/R_outer.

    rng = galsim.BaseDeviate(1234).as_numpy_generator()
    x = rng.uniform(-1.0, 1.0, size=100)
    y = rng.uniform(-1.0, 1.0, size=100)

    R_outer = 1.0
    R_inner = 0.5
    eps = R_inner/R_outer

    test_vals = [
        (10, 1e-12, 1e-12),  # Z66
        (20, 1e-12, 1e-12),  # Z231
        (40, 1e-9, 1e-12),   # Z861
    ]
    if run_slow:
        test_vals += [
            (60, 1e-6, 1e-12),  # Z1891
            (80, 1e-3, 1e-12),  # Z3321
            (100, None, 1e-11),   # Z5151
            (200, None, 1e-11),   # Z20301
        ]

    print()
    for n, tol, tol2 in test_vals:
        j = (n+1)*(n+2)//2
        _, m = galsim.zernike.noll_to_zern(j)
        print(f"Z{j} => (n, m) = ({n}, {m})")
        assert n == abs(m)
        coefs = [0]*j+[1]
        zk = Zernike(coefs, R_outer=R_outer, R_inner=R_inner)

        def analytic_zk(x, y):
            r = np.hypot(x, y)
            theta = np.arctan2(y, x)
            factor = np.sqrt((2*n+2) / np.sum([eps**(2*i) for i in range(n+1)]))
            if m > 0:
                return r**n * np.cos(n*theta) * factor
            else:
                return r**n * np.sin(n*theta) * factor

        analytic_vals = analytic_zk(x, y)
        if n < 100:
            np.testing.assert_allclose(
                zk(x, y),
                analytic_vals,
                atol=tol, rtol=tol
            )

        robust_vals = zk.evalCartesianRobust(x, y)
        np.testing.assert_allclose(
            robust_vals,
            analytic_vals,
            atol=tol2, rtol=tol2
        )
        np.testing.assert_equal(
            robust_vals,
            zk(x, y, robust=True)
        )


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#


import os
import sys
import numpy as np
import time
from unittest import mock

import galsim
from galsim_test_helpers import *

path, filename = os.path.split(__file__) # Get the path to this file for use below...

TESTDIR=os.path.join(path, "table_comparison_files")

DECIMAL = 14 # Make sure output agrees at 14 decimal places or better

# Some arbitrary args to use for the test:
args1 = range(7)  # Evenly spaced
vals1 = [ x**2 for x in args1 ]
testargs1 = [ 0.1, 0.8, 2.3, 3, 5.6, 5.9 ] # < 0 or > 7 is invalid

args2 = [ 0.7, 3.3, 14.1, 15.6, 29, 34.1, 42.5 ]  # Not evenly spaced
vals2 = [ np.sin(x*np.pi/180) for x in args2 ]
testargs2 = [ 1.1, 10.8, 12.3, 15.6, 25.6, 41.9 ] # < 0.7 or > 42.5 is invalid

interps = [ 'linear', 'spline', 'floor', 'ceil', 'nearest' ]


@timer
def test_table():
    """Test the spline tabulation of the k space Cubic interpolant.

    Compares interpolated values a LookupTable that were created using a previous version of
    the code (commit: e267f058351899f1f820adf4d6ab409d5b2605d5), using the
    script devutils/external/make_table_testarrays.py
    """
    for interp in interps:
        table1 = galsim.LookupTable(x=args1,f=vals1,interpolant=interp)
        testvals1 = [ table1(x) for x in testargs1 ]
        assert len(table1) == len(args1)

        np.testing.assert_array_equal(table1.getArgs(), args1)
        np.testing.assert_array_equal(table1.getVals(), vals1)
        assert table1.interpolant == interp
        assert table1.isLogX() == False
        assert table1.isLogF() == False

        # The 4th item is in the args list, so it should be exactly the same as the
        # corresponding item in the vals list.
        np.testing.assert_almost_equal(testvals1[3], vals1[3], DECIMAL,
                err_msg="Interpolated value for exact arg entry does not match val entry")

        # Compare the results in testvals1 with the results if we reshape the input array to be
        # 2-dimensional.
        np.testing.assert_array_almost_equal(
            np.array(testvals1).reshape((2,3)), table1(np.array(testargs1).reshape((2,3))),
            DECIMAL,
            err_msg="Interpolated values do not match when input array shape changes")

        if interp != 'nearest':
            # Do a full regression test based on a version of the code thought to be working.
            ref1 = np.loadtxt(os.path.join(TESTDIR, 'table_test1_%s.txt'%interp))
            np.testing.assert_array_almost_equal(ref1, testvals1, DECIMAL,
                    err_msg="Interpolated values from LookupTable do not match saved "+
                    "data for evenly-spaced args, with interpolant %s."%interp)

            # Same thing, but now for args that are not evenly spaced.
            # (The Table class uses a different algorithm when the arguments are evenly spaced
            #  than when they are not.)
            table2 = galsim.LookupTable(x=args2,f=vals2,interpolant=interp)
            testvals2 = [ table2(x) for x in testargs2 ]

            np.testing.assert_almost_equal(testvals2[3], vals2[3], DECIMAL,
                    err_msg="Interpolated value for exact arg entry does not match val entry")
            ref2 = np.loadtxt(os.path.join(TESTDIR, 'table_test2_%s.txt'%interp))
            np.testing.assert_array_almost_equal(ref2, testvals2, DECIMAL,
                    err_msg="Interpolated values from LookupTable do not match saved "+
                    "data for non-evenly-spaced args, with interpolant %s."%interp)

        # Check that out of bounds arguments, or ones with some crazy shape, raise an exception:
        assert_raises(ValueError,table1,args1[0]-0.01)
        assert_raises(ValueError,table1,args1[-1]+0.01)
        assert_raises(ValueError,table2,args2[0]-0.01)
        assert_raises(ValueError,table2,args2[-1]+0.01)

        # These shouldn't raise any exception:
        table1(args1[0]+0.01)
        table1(args1[-1]-0.01)
        table2(args2[0]+0.01)
        table2(args2[-1]-0.01)
        table1(np.zeros((3,3))+args1[0]+0.01)
        table1(np.zeros(3)+args1[0]+0.01)
        table1((args1[0]+0.01,args1[0]+0.01))
        table1([args1[0]+0.01,args1[0]+0.01])
        # Check 2d arrays (square)
        table1(np.zeros((3,3))+args1[0])
        # Check 2d arrays (non-square)
        table1(np.array(testargs1).reshape((2,3)))

        # Check picklability
        check_pickle(table1, lambda x: (tuple(x.getArgs()), tuple(x.getVals()), x.getInterp()))
        check_pickle(table2, lambda x: (tuple(x.getArgs()), tuple(x.getVals()), x.getInterp()))
        check_pickle(table1)
        check_pickle(table2)

        table3 = galsim._LookupTable(x=args1,f=vals1,interpolant=interp)
        assert table3 == table1

    # Spline interpolation with only 2 points is just linear.
    table1 = galsim.LookupTable([1,5],[9,3],'linear')
    table2 = galsim.LookupTable([1,5],[9,3],'spline')
    assert np.isclose(table1(3), 6)
    assert np.isclose(table2(3), 6)
    assert table1.interpolant == 'linear'
    assert table2.interpolant == 'spline'

    assert_raises(ValueError, galsim.LookupTable, x=args1, f=vals1, interpolant='invalid')
    assert_raises(ValueError, galsim.LookupTable, x=[1], f=[1], interpolant='linear')
    assert_raises(ValueError, galsim.LookupTable, x=[], f=[])
    assert_raises(ValueError, galsim.LookupTable, x=[1,1,1], f=[1,2,3])
    assert_raises(ValueError, galsim.LookupTable, x=[0,1,2], f=[1,2,3], x_log=True)
    assert_raises(ValueError, galsim.LookupTable, x=[-1,0,1], f=[1,2,3], x_log=True)
    assert_raises(ValueError, galsim.LookupTable, x=[0,1,2], f=[0,1,2], f_log=True)
    assert_raises(ValueError, galsim.LookupTable, x=[0,1,2], f=[2,-1,2], f_log=True)
    assert_raises(ValueError, galsim.LookupTable, x=args2, f=vals2, interpolant=galsim.Linear())
    assert_raises(ValueError, galsim.LookupTable, x=args1, f=vals1, interpolant=galsim.Linear(),
                  x_log=True)


@timer
def test_init():
    """Some simple tests of LookupTable initialization."""

    # Make sure nothing bad happens when we try to read in a stored power spectrum and assume
    # we can use the default interpolant (spline).
    tab_ps = galsim.LookupTable.from_file('../examples/data/cosmo-fid.zmed1.00_smoothed.out')
    check_pickle(tab_ps)

    # Check for bad inputs
    assert_raises(TypeError, galsim.LookupTable, x='foo')
    assert_raises(TypeError, galsim.LookupTable)
    assert_raises(TypeError, galsim.LookupTable, x=tab_ps.x)
    assert_raises(TypeError, galsim.LookupTable, f=tab_ps.f)
    assert_raises(ValueError, galsim.LookupTable, x=tab_ps.x, f=tab_ps.f, interpolant='foo')

    with assert_raises(ValueError):
        # This file is for a different purpose. It has one column, which is invalid here.
        galsim.LookupTable.from_file('table_comparison_files/table_test1_spline.txt')

    # The default reader uses pandas, but numpy is used if pandas not available.
    with mock.patch.dict(sys.modules, {'pandas':None}):
        tab_ps2 = galsim.LookupTable.from_file('../examples/data/cosmo-fid.zmed1.00_smoothed.out')
        # Not precisely equal. But close enough.
        np.testing.assert_allclose(tab_ps2.x, tab_ps.x, rtol=1.e-12, atol=1.e-12)
        np.testing.assert_allclose(tab_ps2.f, tab_ps.f, rtol=1.e-12, atol=1.e-12)

    # We expect that people will normally initialize with sorted x values, but that's not actually
    # required.  Check that we get the smae thing with a different order.
    index = np.arange(len(tab_ps.x))
    np.random.shuffle(index)
    x1 = tab_ps.x[index]
    f1 = tab_ps.f[index]
    tab_ps3 = galsim.LookupTable(x1, f1)
    assert tab_ps3 == tab_ps

@timer
def test_log():
    """Some simple tests of interpolation using logs."""
    # Set up some test vectors that are strictly positive, and others that are negative.
    x = 0.01*np.arange(1000)+0.01
    y = 1.*x
    x_neg = -1.*x
    y_neg = 1.*x_neg

    # Check that interpolation agrees for the positive ones when using log interpolation (for some
    # reasonable tolerance).
    tab_1 = galsim.LookupTable(x=x, f=y)
    tab_2 = galsim.LookupTable(x=x, f=y, x_log=True, f_log=True)
    tab_3 = galsim.LookupTable(x=x, f=y, x_log=True)
    tab_4 = galsim.LookupTable(x=x, f=y, f_log=True)
    test_x_vals = [2.641, 3.985, 8.123125]
    for test_val in test_x_vals:
        result_1 = tab_1(test_val)
        result_2 = tab_2(test_val)
        result_3 = tab_3(test_val)
        result_4 = tab_4(test_val)
        print(result_1, result_2, result_3, result_4)
        np.testing.assert_almost_equal(
            result_2, result_1, decimal=3,
            err_msg='Disagreement when interpolating in log(f) and log(x)')
        np.testing.assert_almost_equal(
            result_3, result_1, decimal=3,
            err_msg='Disagreement when interpolating in log(x)')
        np.testing.assert_almost_equal(
            result_4, result_1, decimal=3,
            err_msg='Disagreement when interpolating in log(f)')

    # Before getting to the range check, numpy will issue a warning about
    # "invalid value encountered in log".
    # Seems like ok additional information for end users.  But we want to ignore that here.
    with assert_raises(galsim.GalSimRangeError):
        with assert_warns(RuntimeWarning):
            tab_2(-1)
    with assert_raises(galsim.GalSimRangeError):
        with assert_warns(RuntimeWarning):
            tab_3(-1)
    with assert_raises(galsim.GalSimRangeError):
        with assert_warns(RuntimeWarning):
            tab_2(x_neg)
    with assert_raises(galsim.GalSimRangeError):
        with assert_warns(RuntimeWarning):
            tab_3(x_neg)

    # Check picklability
    check_pickle(tab_1)
    check_pickle(tab_2)
    check_pickle(tab_3)
    check_pickle(tab_4)

    # Check storage of args and vals for log vs. linear, which should be the same to high precision.
    np.testing.assert_array_almost_equal(tab_1.getArgs(), tab_3.getArgs(), decimal=12,
                                         err_msg='Args differ for linear vs. log storage')
    np.testing.assert_array_almost_equal(tab_1.getVals(), tab_4.getVals(), decimal=12,
                                         err_msg='Vals differ for linear vs. log storage')
    # Check other properties
    assert not tab_1.x_log
    assert not tab_1.f_log
    assert tab_2.x_log
    assert tab_2.f_log
    assert tab_3.x_log
    assert not tab_3.f_log
    assert not tab_1.isLogX()
    assert not tab_1.isLogF()
    assert tab_2.isLogX()
    assert tab_2.isLogF()
    assert tab_3.isLogX()
    assert not tab_3.isLogF()

    # Check that an appropriate exception is thrown when trying to do interpolation using negative
    # ones.
    assert_raises(ValueError, galsim.LookupTable, x=x_neg, f=y_neg, x_log=True)
    assert_raises(ValueError, galsim.LookupTable, x=x_neg, f=y_neg, f_log=True)
    assert_raises(ValueError, galsim.LookupTable, x=x_neg, f=y_neg, x_log=True, f_log=True)

    # Check that doing log transform explicitly matches expected behavior of x_log and f_log.
    expx = np.exp(x)
    expy = np.exp(y)
    for interpolant in ['linear', 'spline', galsim.Quintic()]:
        tab_1 = galsim.LookupTable(x, y, interpolant=interpolant)
        tab_2 = galsim.LookupTable(expx, y, x_log=True, interpolant=interpolant)
        tab_3 = galsim.LookupTable(x, expy, f_log=True, interpolant=interpolant)
        tab_4 = galsim.LookupTable(expx, expy, x_log=True, f_log=True, interpolant=interpolant)
        tab_5 = galsim._LookupTable(expx, expy, x_log=True, f_log=True, interpolant=interpolant)
        for test_val in test_x_vals:
            result_1 = tab_1(test_val)
            result_2 = tab_2(np.exp(test_val))
            result_3 = np.log(tab_3(test_val))
            result_4 = np.log(tab_4(np.exp(test_val)))
            result_5 = np.log(tab_5(np.exp(test_val)))
            np.testing.assert_almost_equal(
                result_2, result_1, decimal=10,
                err_msg='Disagreement when interpolating in log(x)')
            np.testing.assert_almost_equal(
                result_3, result_1, decimal=10,
                err_msg='Disagreement when interpolating in log(f)')
            np.testing.assert_almost_equal(
                result_4, result_1, decimal=10,
                err_msg='Disagreement when interpolating in log(f) and log(x)')
            np.testing.assert_almost_equal(
                result_5, result_1, decimal=10,
                err_msg='Disagreement when interpolating in log(f) and log(x) with _LookupTable')

    # Verify for exception when using x_log with non-equal-spaced galsim.Interpolant
    galsim.LookupTable(x, y, x_log=True, interpolant='linear') # works fine
    assert_raises(ValueError, galsim.LookupTable, x, y, x_log=True, interpolant=galsim.Linear())


@timer
def test_from_func():
    """Test the LookupTable.from_func factory function"""
    x_min = 2
    x_max = 200

    # Linear interpolation
    x1 = np.linspace(x_min, x_max, 2000)
    f1 = [x**3 for x in x1]
    tab1 = galsim.LookupTable(x1, f1, interpolant='linear')
    tab2 = galsim.LookupTable.from_func(lambda x:x**3, x_min, x_max, interpolant='linear')
    print('tab1 = ',tab1, tab1(10))
    print('tab2 = ',tab2, tab2(10))

    # Spline interpolation
    tab3 = galsim.LookupTable(x1, f1)
    tab4 = galsim.LookupTable.from_func(lambda x:x**3, x_min, x_max)
    print('tab3 = ',tab3, tab3(10))
    print('tab4 = ',tab4, tab4(10))

    # Log interpolation
    x5 = np.exp(np.linspace(np.log(x_min), np.log(x_max), 2000))
    f5 = [x**3 for x in x5]
    tab5 = galsim.LookupTable(x5, f5, x_log=True, f_log=True)
    tab6 = galsim.LookupTable.from_func(lambda x:x**3, x_min, x_max, x_log=True, f_log=True)
    print('tab5 = ',tab5, tab5(10))
    print('tab6 = ',tab6, tab6(10))

    test_x_vals = [2.641, 39.85, 81.23125]
    for x in test_x_vals:
        truth = x**3
        f1 = tab1(x)
        f2 = tab2(x)
        f3 = tab3(x)
        f4 = tab4(x)
        f5 = tab5(x)
        f6 = tab6(x)
        print(truth, f1, f2, f3, f4, f5, f6)
        np.testing.assert_almost_equal(f1/truth, 1.0, decimal=2)
        np.testing.assert_almost_equal(f2/truth, 1.0, 2,
                                       "LookupTable.from_func (linear) gave wrong answer")
        np.testing.assert_almost_equal(f3/truth, 1.0, decimal=6)
        np.testing.assert_almost_equal(f4/truth, 1.0, 6,
                                       "LookupTable.from_func (spline) gave wrong answer")
        np.testing.assert_almost_equal(f5/truth, 1.0, decimal=11)
        np.testing.assert_almost_equal(f6/truth, 1.0, 11,
                                       "LookupTable.from_func (log-log) gave wrong answer")
    check_pickle(tab2)
    check_pickle(tab4)
    check_pickle(tab6)


@timer
def test_roundoff():
    table1 = galsim.LookupTable([1,2,3,4,5,6,7,8,9,10], [1,2,3,4,5,6,7,8,9,10])
    # These should work without raising an exception
    np.testing.assert_almost_equal(table1(1.0 - 1.e-7), 1.0, decimal=6)
    np.testing.assert_almost_equal(table1(10.0 + 1.e-7), 10.0, decimal=6)
    assert_raises(ValueError, table1, 1.0-1.e5)
    assert_raises(ValueError, table1, 10.0+1.e5)


@timer
def test_table_GSInterp():
    def f(x_):
        return x_ + 2*x_*x_ + 3*np.sin(2*x_)

    x = np.linspace(0.1, 3.3, 25)
    y = np.arange(20) # Need something big enough to avoid having the interpolant fall off the edge
    xx, yy = np.meshgrid(x, y)

    interpolants = ['lanczos3', 'lanczos3F', 'lanczos7', 'sinc', 'quintic']

    for interpolant in interpolants:
        z = f(xx)

        tab = galsim.LookupTable(x, z[0,:], interpolant=interpolant)
        check_pickle(tab)

        # Use InterpolatedImage to validate
        wcs = galsim.JacobianWCS(
            (max(x) - min(x))/(len(x)-1),
            0, 0,
            (max(y) - min(y))/(len(y)-1),
        )
        img = galsim.Image(z, wcs=wcs)
        ii = galsim.InterpolatedImage(
            img,
            use_true_center=True,
            offset=(galsim.PositionD(img.xmin,img.ymin)-img.true_center),
            x_interpolant=interpolant,
            normalization='sb',
            calculate_maxk=False, calculate_stepk=False
        ).shift(min(x), min(y))

        # Check single value functionality.
        x1 = 2.3
        np.testing.assert_allclose(tab(x1), ii.xValue(x1, 10), atol=1e-10, rtol=0)

        # Check vectorized output
        newx = np.linspace(0.2, 3.1, 15)
        np.testing.assert_allclose(
            tab(newx),
            np.array([ii.xValue(x_, 10) for x_ in newx]),
            atol=1e-10, rtol=0
        )

        tab2 = galsim._LookupTable(x, z[0,:], interpolant=interpolant)
        assert tab2 == tab


@timer
def test_table2d():
    """Check LookupTable2D functionality.
    """
    from scipy.interpolate import RectBivariateSpline

    def f(x_, y_):
        return np.sin(x_) * np.cos(y_) + x_

    x = np.linspace(0.1, 3.3, 25)
    y = np.linspace(0.2, 10.4, 75)
    xx, yy = np.meshgrid(x, y)
    z = f(xx, yy)

    tab2d = galsim.LookupTable2D(x, y, z)
    check_pickle(tab2d)

    np.testing.assert_array_equal(tab2d.getXArgs(), x)
    np.testing.assert_array_equal(tab2d.getYArgs(), y)
    np.testing.assert_array_equal(tab2d.getVals(), z)
    assert tab2d.interpolant == 'linear'
    assert tab2d.edge_mode == 'raise'

    newx = np.linspace(0.2, 3.1, 45)
    newy = np.linspace(0.3, 10.1, 85)
    newxx, newyy = np.meshgrid(newx, newy)

    # Compare different ways of evaluating Table2D
    ref = tab2d(newxx, newyy)
    np.testing.assert_array_almost_equal(ref, tab2d(newx, newy, grid=True))
    np.testing.assert_array_almost_equal(ref, np.array([[tab2d(x0, y0)
                                                         for x0 in newx]
                                                        for y0 in newy]))

    scitab2d = RectBivariateSpline(x, y, z.T, kx=1, ky=1)
    np.testing.assert_array_almost_equal(ref, scitab2d(newx, newy).T)

    # Try using linear GSInterp
    tab2d2 = galsim.LookupTable2D(x, y, z, interpolant=galsim.Linear())
    np.testing.assert_array_almost_equal(tab2d(newxx, newyy), tab2d2(newxx, newyy))
    np.testing.assert_array_almost_equal(tab2d(newxx.T, newyy.T), tab2d2(newxx.T, newyy.T))

    # Try again using Nearest()
    tab2d2 = galsim.LookupTable2D(x, y, z, interpolant=galsim.Nearest())
    tab2d3 = galsim.LookupTable2D(x, y, z, interpolant='nearest')
    np.testing.assert_array_almost_equal(tab2d2(newxx, newyy), tab2d3(newxx, newyy))
    np.testing.assert_array_almost_equal(tab2d2(newxx.T, newyy.T), tab2d3(newxx.T, newyy.T))
    # Make sure we exercise the special case in T2DInterpolant2D::interp
    np.testing.assert_array_almost_equal(tab2d2(0.3, 0.4), tab2d3(0.3, 0.4))
    np.testing.assert_array_almost_equal(tab2d2(0.3, 0.4), tab2d3(0.3, 0.4))


    # Test non-equally-spaced table.
    x = np.delete(x, 10)
    y = np.delete(y, 10)
    xx, yy = np.meshgrid(x, y)
    z = f(xx, yy)
    tab2d = galsim.LookupTable2D(x, y, z)
    ref = tab2d(newxx, newyy)
    np.testing.assert_array_almost_equal(ref, tab2d(newx, newy, grid=True))
    np.testing.assert_array_almost_equal(ref, np.array([[tab2d(x0, y0)
                                                         for x0 in newx]
                                                        for y0 in newy]))
    scitab2d = RectBivariateSpline(x, y, z.T, kx=1, ky=1)
    np.testing.assert_array_almost_equal(ref, scitab2d(newx, newy).T)

    # Using a galsim.Interpolant should raise an exception if x/y are not equal spaced.
    with assert_raises(galsim.GalSimIncompatibleValuesError):
        tab2d2 = galsim.LookupTable2D(x, y, z, interpolant=galsim.Linear())

    # Try a simpler interpolation function.  We should be able to interpolate a (bi-)linear function
    # exactly with a linear interpolant.
    def f(x_, y_):
        return 2*x_ + 3*y_

    z = f(xx, yy)
    tab2d = galsim.LookupTable2D(x, y, z)

    np.testing.assert_array_almost_equal(f(newxx, newyy), tab2d(newxx, newyy))
    np.testing.assert_array_almost_equal(f(newxx, newyy), np.array([[tab2d(x0, y0)
                                                                     for x0 in newx]
                                                                    for y0 in newy]))

    # Test edge exception
    with assert_raises(ValueError):
        tab2d(1e6, 1e6)
    with assert_raises(ValueError):
        tab2d(np.array([1e5, 1e6]), np.array([1e5, 1e6]))
    with assert_raises(ValueError):
        tab2d(np.array([1e5, 1e6]), np.array([1e5, 1e6]), grid=True)
    with assert_raises(ValueError):
        tab2d.gradient(1e6, 1e6)
    with assert_raises(ValueError):
        tab2d.gradient(np.array([1e5, 1e6]), np.array([1e5, 1e6]))
    with assert_raises(ValueError):
        tab2d.gradient(np.array([1e5, 1e6]), np.array([1e5, 1e6]), grid=True)
    with assert_raises(galsim.GalSimError):
        galsim.utilities.find_out_of_bounds_position(
            np.array([1, 3]), np.array([2, 4]), galsim.BoundsD(0,5,0,5))
    with assert_raises(galsim.GalSimError):
        galsim.utilities.find_out_of_bounds_position(
            np.array([1, 3]), np.array([2, 4]), galsim.BoundsD(0,5,0,5), grid=True)

    # Check warning mode
    tab2dw = galsim.LookupTable2D(x, y, z, edge_mode='warn', constant=1)
    with assert_warns(galsim.GalSimWarning):
        assert tab2dw(1e6, 1e6) == 1
    with assert_warns(galsim.GalSimWarning):
        np.testing.assert_array_equal(
            tab2dw(np.array([1e5, 1e6]), np.array([1e5, 1e6])),
            np.array([1.0, 1.0])
        )
    with assert_warns(galsim.GalSimWarning):
        np.testing.assert_array_equal(
            tab2dw(np.array([1e5, 1e6]), np.array([1e5, 1e6]), grid=True),
            np.array([[1.0, 1.0], [1.0, 1.0]])
        )
    with assert_warns(galsim.GalSimWarning):
        assert tab2dw.gradient(1e6, 1e6) == (0.0, 0.0)
    with assert_warns(galsim.GalSimWarning):
        np.testing.assert_array_equal(
            tab2dw.gradient(np.array([1e5, 1e6]), np.array([1e5, 1e6])),
            np.zeros((2, 2), dtype=float)
        )
    with assert_warns(galsim.GalSimWarning):
        np.testing.assert_array_equal(
            tab2dw.gradient(np.array([1e5, 1e6]), np.array([1e5, 1e6]), grid=True),
            np.zeros((2, 2, 2), dtype=float)
        )

    # But doesn't warn if in bounds
    tab2dw(1.0, 1.0)
    tab2dw(np.array([1.0]), np.array([1.0]))
    tab2dw(np.array([1.0]), np.array([1.0]), grid=True)
    tab2dw.gradient(1.0, 1.0)
    tab2dw.gradient(np.array([1.0]), np.array([1.0]))

    # Test edge wrapping
    # Check that can't construct table with edge-wrapping if edges don't match
    with assert_raises(ValueError):
        galsim.LookupTable2D(x, y, z, edge_mode='wrap')

    # Extend edges and make vals match
    x = np.append(x, x[-1] + (x[-1]-x[-2]))
    y = np.append(y, y[-1] + (y[-1]-y[-2]))
    z = np.pad(z,[(0,1), (0,1)], mode='wrap')
    tab2d = galsim.LookupTable2D(x, y, z, edge_mode='wrap')

    np.testing.assert_array_almost_equal(tab2d(newxx, newyy), tab2d(newxx+3*(x[-1]-x[0]), newyy))
    np.testing.assert_array_almost_equal(
        tab2d(newx, newy, grid=True), tab2d(newxx+3*(x[-1]-x[0]), newyy))
    np.testing.assert_array_almost_equal(tab2d(newxx, newyy), tab2d(newxx, newyy+13*(y[-1]-y[0])))
    np.testing.assert_array_almost_equal(
        tab2d(newxx, newyy), tab2d(newx, newy+13*(y[-1]-y[0]), grid=True))

    # Test edge_mode='constant'
    tab2d = galsim.LookupTable2D(x, y, z, edge_mode='constant', constant=42)
    assert type(tab2d(x[0]-1, y[0]-1)) in [float, np.float64]
    assert tab2d(x[0]-1, y[0]-1) == 42.0
    # One in-bounds, one out-of-bounds
    np.testing.assert_array_almost_equal(tab2d([x[0], x[0]-1], [y[0], y[0]-1]),
                                         [tab2d(x[0], y[0]), 42.0])


    # Test floor/ceil/nearest interpolant
    x = np.arange(5)
    y = np.arange(5)
    z = x + y[:, np.newaxis]
    tab2d = galsim.LookupTable2D(x, y, z, interpolant='ceil')
    assert tab2d(2.4, 3.6) == 3+4, "Ceil interpolant failed."
    tab2d = galsim.LookupTable2D(x, y, z, interpolant='floor')
    assert tab2d(2.4, 3.6) == 2+3, "Floor interpolant failed."
    tab2d = galsim.LookupTable2D(x, y, z, interpolant='nearest')
    assert tab2d(2.4, 3.6) == 2+4, "Nearest interpolant failed."
    tab2d = galsim.LookupTable2D(x, y, z, interpolant=galsim.Nearest())
    assert tab2d(2.4, 3.6) == 2+4, "Nearest interpolant failed."

    assert_raises(ValueError, galsim.LookupTable2D, x, y, z, interpolant='invalid')
    assert_raises(ValueError, galsim.LookupTable2D, x, y, z, edge_mode='invalid')
    assert_raises(ValueError, galsim.LookupTable2D, x, y, z[:-1,:-1])

    # Test that x,y arrays need to be strictly increasing.
    x[0] = x[1]
    assert_raises(ValueError, galsim.LookupTable2D, x, y, z)
    x[0] = x[1]+1
    assert_raises(ValueError, galsim.LookupTable2D, x, y, z)
    x[0] = x[1]-1
    y[0] = y[1]
    assert_raises(ValueError, galsim.LookupTable2D, x, y, z)
    y[0] = y[1]+1
    assert_raises(ValueError, galsim.LookupTable2D, x, y, z)

    # Test that x,y arrays are larger than interpolant
    x = y = np.arange(2)
    z = x[:,None]+y
    assert_raises(ValueError, galsim.LookupTable2D, x, y, z, interpolant='spline', edge_mode='wrap')

    # Check dfdx input
    x = y = np.arange(20)
    z = x[:,None]+y
    # Not using interpolant='spline'
    assert_raises(ValueError, galsim.LookupTable2D, x, y, z, dfdx=z)
    # Only specifying one derivative
    assert_raises(ValueError, galsim.LookupTable2D, x, y, z, dfdx=z, interpolant='spline')
    # Derivative is wrong shape
    assert_raises(ValueError, galsim.LookupTable2D, x, y, z, dfdx=z, dfdy=z, d2fdxdy=z[::2],
                  interpolant='spline')

    # Check private shortcut instantiation
    new_tab2d = galsim.table._LookupTable2D(
        tab2d.x, tab2d.y, tab2d.f,
        tab2d.interpolant, tab2d.edge_mode,
        tab2d.constant, tab2d.dfdx, tab2d.dfdy, tab2d.d2fdxdy
    )
    assert tab2d == new_tab2d
    assert tab2d(2.4, 3.6) == new_tab2d(2.4, 3.6)


@timer
def test_table2d_gradient():
    """Check LookupTable2D gradient function
    """
    # Same function as the above test
    def f(x_, y_):
        return np.sin(x_) * np.cos(y_) + x_
    # The gradient is analytic for this:
    def dfdx(x_, y_):
        return np.cos(x_) * np.cos(y_) + 1.
    def dfdy(x_, y_):
        return -np.sin(x_) * np.sin(y_)

    x = np.linspace(0.1, 3.3, 250)
    y = np.linspace(0.2, 10.4, 750)
    xx, yy = np.meshgrid(x, y)
    z = f(xx, yy)

    tab2d = galsim.LookupTable2D(x, y, z)

    newx = np.linspace(0.2, 3.1, 45)
    newy = np.linspace(0.3, 10.1, 85)
    newxx, newyy = np.meshgrid(newx, newy)

    # Check single value functionality.
    x1,y1 = 1.1, 4.9
    np.testing.assert_almost_equal(tab2d.gradient(x1,y1), (dfdx(x1,y1), dfdy(x1,y1)), decimal=2)

    # Check that the gradient function comes out close to the analytic derivatives.
    ref_dfdx = dfdx(newxx, newyy)
    ref_dfdy = dfdy(newxx, newyy)
    test_dfdx, test_dfdy = tab2d.gradient(newxx, newyy)
    np.testing.assert_almost_equal(test_dfdx, ref_dfdx, decimal=2)
    np.testing.assert_almost_equal(test_dfdy, ref_dfdy, decimal=2)
    test2_dfdx, test2_dfdy = tab2d.gradient(newx, newy, grid=True)
    np.testing.assert_almost_equal(test2_dfdx, ref_dfdx, decimal=2)
    np.testing.assert_almost_equal(test2_dfdy, ref_dfdy, decimal=2)

    # Check edge wrapping
    tab2d = galsim.LookupTable2D(x, y, z, edge_mode='wrap')

    test_dfdx, test_dfdy = tab2d.gradient(newxx+13*tab2d.xperiod, newyy)
    test2_dfdx, test2_dfdy = tab2d.gradient(newx+13*tab2d.xperiod, newy, grid=True)
    np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx, decimal=2)
    np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy, decimal=2)
    np.testing.assert_array_almost_equal(ref_dfdx, test2_dfdx, decimal=2)
    np.testing.assert_array_almost_equal(ref_dfdy, test2_dfdy, decimal=2)

    test_dfdx, test_dfdy = tab2d.gradient(newxx, newyy+12*tab2d.yperiod)
    np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx, decimal=2)
    np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy, decimal=2)

    # Test single value:
    np.testing.assert_almost_equal(tab2d.gradient(x1, y1), tab2d.gradient(x1-9*tab2d.xperiod, y1))
    np.testing.assert_almost_equal(tab2d.gradient(x1, y1), tab2d.gradient(x1, y1-7*tab2d.yperiod))


    # Check constant edge_mode
    tab2d = galsim.LookupTable2D(x, y, z, edge_mode='constant')

    # Should work the same inside original boundary
    test_dfdx, test_dfdy = tab2d.gradient(newxx, newyy)
    test2_dfdx, test2_dfdy = tab2d.gradient(newx, newy, grid=True)
    np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx, decimal=2)
    np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy, decimal=2)
    np.testing.assert_array_almost_equal(ref_dfdx, test2_dfdx, decimal=2)
    np.testing.assert_array_almost_equal(ref_dfdy, test2_dfdy, decimal=2)

    # Should come out zero outside original boundary
    test_dfdx, test_dfdy = tab2d.gradient(newxx+2*np.max(newxx), newyy+2*np.max(newyy))
    test2_dfdx, test2_dfdy = tab2d.gradient(newx+2*np.max(newx), newy+2*np.max(newy), grid=True)
    np.testing.assert_array_equal(0.0, test_dfdx)
    np.testing.assert_array_equal(0.0, test_dfdy)
    np.testing.assert_array_equal(0.0, test2_dfdx)
    np.testing.assert_array_equal(0.0, test2_dfdy)

    # Test single value
    np.testing.assert_almost_equal(tab2d.gradient(x1, y1), (dfdx(x1,y1), dfdy(x1, y1)), decimal=2)
    np.testing.assert_equal(tab2d.gradient(x1+2*np.max(newxx), y1), (0.0, 0.0))


    # Try a simpler interpolation function.  Derivatives should be exact.
    def f(x_, y_):
        return 2*x_ + 3*y_ - 4*x_*y_

    z = f(xx, yy)

    tab2d = galsim.LookupTable2D(x, y, z)
    test_dfdx, test_dfdy = tab2d.gradient(newxx, newyy)
    test2_dfdx, test2_dfdy = tab2d.gradient(newx, newy, grid=True)
    np.testing.assert_almost_equal(test_dfdx, 2.-4*newyy, decimal=7)
    np.testing.assert_almost_equal(test_dfdy, 3.-4*newxx, decimal=7)
    np.testing.assert_almost_equal(test2_dfdx, 2.-4*newyy, decimal=7)
    np.testing.assert_almost_equal(test2_dfdy, 3.-4*newxx, decimal=7)

    # Check single value functionality.
    np.testing.assert_almost_equal(tab2d.gradient(x1,y1), (2.-4*y1, 3.-4*x1))

    # Check edge wrapping
    tab2d = galsim.LookupTable2D(x, y, z, edge_mode='wrap')

    ref_dfdx, ref_dfdy = tab2d.gradient(newxx, newyy)
    test_dfdx, test_dfdy = tab2d.gradient(newxx+3*tab2d.xperiod, newyy)
    test2_dfdx, test2_dfdy = tab2d.gradient(newx+3*tab2d.xperiod, newy, grid=True)
    np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx)
    np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy)
    np.testing.assert_array_almost_equal(ref_dfdx, test2_dfdx)
    np.testing.assert_array_almost_equal(ref_dfdy, test2_dfdy)

    test_dfdx, test_dfdy = tab2d.gradient(newxx, newyy+13*tab2d.yperiod)
    np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx)
    np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy)

    # Test single value
    np.testing.assert_almost_equal(tab2d.gradient(x1, y1), (2-4*y1, 3-4*x1))
    np.testing.assert_almost_equal(tab2d.gradient(x1+2*tab2d.xperiod, y1), (2-4*y1, 3-4*x1))

    # Test mix of inside and outside original boundary
    test_dfdx, test_dfdy = tab2d.gradient(np.dstack([newxx, newxx+3*tab2d.xperiod]),
                                          np.dstack([newyy, newyy]))

    np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx[:,:,0])
    np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy[:,:,0])
    np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx[:,:,1])
    np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy[:,:,1])

    # Check that edge_mode='constant' behaves as expected.

    # Should work the same inside original boundary
    tab2d = galsim.LookupTable2D(x, y, z, edge_mode='constant')
    test_dfdx, test_dfdy = tab2d.gradient(newxx, newyy)
    np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx)
    np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy)

    # Should come out zero outside original boundary
    test_dfdx, test_dfdy = tab2d.gradient(newxx+2*np.max(newxx), newyy)
    np.testing.assert_array_equal(0.0, test_dfdx)
    np.testing.assert_array_equal(0.0, test_dfdy)

    # Test single value
    np.testing.assert_almost_equal(tab2d.gradient(x1, y1), (2-4*y1, 3-4*x1))
    np.testing.assert_equal(tab2d.gradient(x1+2*np.max(newxx), y1), (0.0, 0.0))

    # Test mix of inside and outside original boundary
    test_dfdx, test_dfdy = tab2d.gradient(np.dstack([newxx, newxx+2*np.max(newxx)]),
                                          np.dstack([newyy, newyy]))

    np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx[:,:,0])
    np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy[:,:,0])
    np.testing.assert_array_equal(0.0, test_dfdx[:,:,1])
    np.testing.assert_array_equal(0.0, test_dfdy[:,:,1])


@timer
def test_table2d_cubic():
    # A few functions that should be exactly interpolatable with bicubic
    # interpolation
    def f1(x_, y_):
        return 2*x_ + 3*y_
    def df1dx(x_, y_):
        return np.ones_like(x_)*2
    def df1dy(x_, y_):
        return np.ones_like(x_)*3

    def f2(x_, y_):
        return 2*x_*x_
    def df2dx(x_, y_):
        return 4*x_
    def df2dy(x_, y_):
        return np.zeros_like(x_)

    def f3(x_, y_):
        return 2*y_*y_
    def df3dx(x_, y_):
        return np.zeros_like(x_)
    def df3dy(x_, y_):
        return 4*y_

    def f4(x_, y_):
        return 2*y_*y_ + 3*x_*x_
    def df4dx(x_, y_):
        return 6*x_
    def df4dy(x_, y_):
        return 4*y_

    def f5(x_, y_):
        return 2*y_*y_ + 3*x_*x_ + 4*x_*y_
    def df5dx(x_, y_):
        return 6*x_ + 4*y_
    def df5dy(x_, y_):
        return 4*y_ + 4*x_

    fs = [f1, f2, f3, f4, f5]
    dfdxs = [df1dx, df2dx, df3dx, df4dx, df5dx]
    dfdys = [df1dy, df2dy, df3dy, df4dy, df5dy]

    x = np.linspace(0.1, 3.3, 250)
    y = np.linspace(0.2, 10.4, 750)
    xx, yy = np.meshgrid(x, y)

    for f, dfdx, dfdy in zip(fs, dfdxs, dfdys):
        z = f(xx, yy)
        tab2d = galsim.LookupTable2D(x, y, z, interpolant='spline')

        # Check single value functionality.
        x1,y1 = 2.3, 1.2
        np.testing.assert_allclose(tab2d(x1,y1), f(x1, y1), atol=1e-11, rtol=0)
        ref_dfdx = dfdx(x1,y1)
        ref_dfdy = dfdy(x1,y1)
        test_dfdx, test_dfdy = tab2d.gradient(x1,y1)
        np.testing.assert_allclose(test_dfdx, ref_dfdx, atol=1e-11, rtol=0)
        np.testing.assert_allclose(test_dfdy, ref_dfdy, atol=1e-11, rtol=0)

        # Check vectorized output
        newx = np.linspace(0.2, 3.1, 45)
        newy = np.linspace(0.3, 10.1, 85)
        newxx, newyy = np.meshgrid(newx, newy)

        np.testing.assert_allclose(tab2d(newxx, newyy), f(newxx, newyy), atol=1e-11, rtol=0)
        ref_dfdx = dfdx(newxx, newyy)
        ref_dfdy = dfdy(newxx, newyy)
        test_dfdx, test_dfdy = tab2d.gradient(newxx, newyy)
        np.testing.assert_allclose(test_dfdx, ref_dfdx, atol=1e-11, rtol=0)
        np.testing.assert_allclose(test_dfdy, ref_dfdy, atol=1e-11, rtol=0)

    # I think it ought to be possible to exactly interpolate the following,
    # provided the exact derivatives are provided.  Use slightly looser tolerances,
    # which seems reasonable since these are varying more rapidly.
    def f6(x_, y_):
        return y_*y_*y_
    def df6dx(x_, y_):
        return np.zeros_like(x_)
    def df6dy(x_, y_):
        return 3*y_*y_
    def d2f6dxdy(x_, y_):
        return np.zeros_like(x_)

    def f7(x_, y_):
        return 2*y_*y_*y_ + 3*x_*x_*x_ + 4*x_*y_*y_
    def df7dx(x_, y_):
        return 9*x_*x_ + 4*y_*y_
    def df7dy(x_, y_):
        return 6*y_*y_ + 8*x_*y_
    def d2f7dxdy(x_, y_):
        return 8*y_

    fs = [f6, f7]
    dfdxs = [df6dx, df7dx]
    dfdys = [df6dy, df7dy]
    d2fdxdys = [d2f6dxdy, d2f7dxdy]

    for f, dfdx, dfdy, d2fdxdy in zip(fs, dfdxs, dfdys, d2fdxdys):
        z = f(xx, yy)
        tab2d = galsim.LookupTable2D(x, y, z, interpolant='spline',
            dfdx=dfdx(xx, yy), dfdy=dfdy(xx, yy), d2fdxdy=d2fdxdy(xx, yy))

        # Check single value functionality.
        x1,y1 = 2.3, 1.2
        np.testing.assert_allclose(tab2d(x1,y1), f(x1, y1), atol=1e-10, rtol=0)
        ref_dfdx = dfdx(x1,y1)
        ref_dfdy = dfdy(x1,y1)
        test_dfdx, test_dfdy = tab2d.gradient(x1,y1)
        np.testing.assert_allclose(test_dfdx, ref_dfdx, atol=1e-10, rtol=0)
        np.testing.assert_allclose(test_dfdy, ref_dfdy, atol=1e-10, rtol=0)

        # Check vectorized output
        newx = np.linspace(0.2, 3.1, 45)
        newy = np.linspace(0.3, 10.1, 85)
        newxx, newyy = np.meshgrid(newx, newy)

        np.testing.assert_allclose(tab2d(newxx, newyy), f(newxx, newyy), atol=1e-10, rtol=0)
        ref_dfdx = dfdx(newxx, newyy)
        ref_dfdy = dfdy(newxx, newyy)
        test_dfdx, test_dfdy = tab2d.gradient(newxx, newyy)
        np.testing.assert_allclose(test_dfdx, ref_dfdx, atol=1e-10, rtol=0)
        np.testing.assert_allclose(test_dfdy, ref_dfdy, atol=1e-10, rtol=0)


@timer
def test_table2d_GSInterp():
    def f(x_, y_):
        return 2*y_*y_ + 3*x_*x_ + 4*x_*y_ - np.cos(x_)

    x = np.linspace(0.1, 3.3, 25)
    y = np.linspace(0.2, 10.4, 75)
    xx, yy = np.meshgrid(x, y)

    interpolants = ['lanczos3', 'lanczos3F', 'lanczos7', 'sinc', 'quintic']

    for interpolant in interpolants:
        z = f(xx, yy)
        tab2d = galsim.LookupTable2D(x, y, z, interpolant=interpolant)
        check_pickle(tab2d)
        # Make sure precomputed-hash gets covered
        hash(tab2d)

        # Use InterpolatedImage to validate
        wcs = galsim.JacobianWCS(
            (max(x) - min(x))/(len(x)-1),
            0, 0,
            (max(y) - min(y))/(len(y)-1),
        )
        img = galsim.Image(z, wcs=wcs)
        ii = galsim.InterpolatedImage(
            img,
            use_true_center=True,
            offset=(galsim.PositionD(img.xmin,img.ymin)-img.true_center),
            x_interpolant=interpolant,
            normalization='sb',
            calculate_maxk=False, calculate_stepk=False
        ).shift(min(x), min(y))

        # Check single value functionality.
        x1,y1 = 2.3, 3.2
        np.testing.assert_allclose(tab2d(x1,y1), ii.xValue(x1, y1), atol=1e-10, rtol=0)

        # Check vectorized output
        newx = np.linspace(0.2, 3.1, 15)
        newy = np.linspace(0.3, 10.1, 25)
        newxx, newyy = np.meshgrid(newx, newy)
        np.testing.assert_allclose(
            tab2d(newxx, newyy).ravel(),
            np.array([ii.xValue(x_, y_)
                      for x_, y_ in zip(newxx.ravel(), newyy.ravel())]).ravel(),
            atol=1e-10, rtol=0
        )
        np.testing.assert_array_almost_equal(tab2d(newxx, newyy), tab2d(newx, newy, grid=True))

        # Check that edge_mode='wrap' works
        tab2d = galsim.LookupTable2D(x, y, z, edge_mode='wrap')

        ref_dfdx, ref_dfdy = tab2d.gradient(newxx, newyy)
        test_dfdx, test_dfdy = tab2d.gradient(newxx+3*tab2d.xperiod, newyy)
        np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx)
        np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy)

        test_dfdx, test_dfdy = tab2d.gradient(newxx, newyy+13*tab2d.yperiod)
        np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx)
        np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy)

        test_dfdx, test_dfdy = tab2d.gradient(newx, newy+13*tab2d.yperiod, grid=True)
        np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx)
        np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy)

        # Test mix of inside and outside original boundary
        test_dfdx, test_dfdy = tab2d.gradient(np.dstack([newxx, newxx+3*tab2d.xperiod]),
                                              np.dstack([newyy, newyy]))

        np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx[:,:,0])
        np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy[:,:,0])
        np.testing.assert_array_almost_equal(ref_dfdx, test_dfdx[:,:,1])
        np.testing.assert_array_almost_equal(ref_dfdy, test_dfdy[:,:,1])


@timer
def test_ne():
    """ Check that inequality works as expected."""
    # These should all compare as unequal.
    x = [1, 2, 3]
    f = [4, 5, 6]
    x2 = [1.1, 2.2, 3.3]
    f2 = [4.4, 5.5, 6.6]
    lts = [galsim.LookupTable(x, f),
           galsim.LookupTable(x, f2),
           galsim.LookupTable(x2, f),
           galsim.LookupTable(x, f, interpolant='floor'),
           galsim.LookupTable(x, f, x_log=True),
           galsim.LookupTable(x, f, f_log=True)]

    ff = f+np.array(f)[:, None]
    ff2 = f2+np.array(f2)[:, None]
    lts.extend([
        galsim.LookupTable2D(x, x, ff),
        galsim.LookupTable2D(x2, x, ff),
        galsim.LookupTable2D(x, x2, ff),
        galsim.LookupTable2D(x, x2, ff2),
        galsim.LookupTable2D(x, x, ff, interpolant='nearest'),
        galsim.LookupTable2D(x, x, ff, interpolant=galsim.Nearest()),
        galsim.LookupTable2D(x, x, ff, edge_mode='wrap'),
        galsim.LookupTable2D(x, x, ff, constant=1),
        galsim.LookupTable2D(x, x, ff, interpolant='spline'),
        galsim.LookupTable2D(x, x, ff, interpolant='spline',
            dfdx=ff, dfdy=ff, d2fdxdy=ff
        )
    ])
    check_all_diff(lts)

@timer
def test_integrate():
    if np.__version__ >= '2.0':
        np_trapz = np.trapezoid
    else:
        np_trapz = np.trapz

    functions = [
        galsim.LookupTable([0,1,2,3,4], [1,5,5,8,1], interpolant='linear'),
        galsim.LookupTable([0,4], [1,8], interpolant='linear'),
        galsim.LookupTable([0,0.1,0.2,3.5,4], [1,5,5,8,1], interpolant='linear'),
        galsim.LookupTable([0,2.1,2.2,2.5,4], [1,5,5,8,1], interpolant='linear'),
        galsim.LookupTable([0,1,2,3,4], [1,5,5,8,1], interpolant='floor'),
        galsim.LookupTable([0,4], [1,8], interpolant='floor'),
        galsim.LookupTable([0,1,2,3,4], [1,5,5,8,1], interpolant='ceil'),
        galsim.LookupTable([0,4], [1,8], interpolant='ceil'),
        galsim.LookupTable([0,1,2,3,4], [1,5,5,8,1], interpolant='nearest'),
        galsim.LookupTable([0,4], [1,8], interpolant='nearest'),
        galsim.LookupTable([0,1,2,3,4], [1,5,5,8,1], interpolant='spline'),
        galsim.LookupTable([0,1,4], [1,1,8], interpolant='spline'),
        galsim.LookupTable([0,0.1,0.2,3.5,4], [1,2,2,8,7], interpolant='spline'),
        galsim.LookupTable([0,2.1,2.2,2.5,4], [1,5,5,5,1], interpolant='spline'),
    ]

    for func in functions:
        print('func = ',repr(func))
        for x_min in [None, func.x_min, func.x_min+1, func.x_min-1]:
            for x_max in [None, func.x_max, func.x_max-1, func.x_max+1]:
                func_ans = func.integrate(x_min, x_max)

                # Note: To keep the original values, x_min, x_max are the loop values
                #       and xmin, xmax are what I use for the test integration, which may
                #       need to be adjusted somewhat.
                if x_min is None or x_min < func.x_min:
                    xmin = func.x_min
                else:
                    xmin = x_min
                if x_max is None or x_max > func.x_max:
                    xmax = func.x_max
                else:
                    xmax = x_max
                x = np.linspace(xmin, xmax, 10000)
                f = func(x)
                np_ans = np_trapz(f,x)
                print('integrate range %s..%s = %s  %s'%(xmin,xmax,func_ans,np_ans))
                if func.interpolant in ['linear', 'spline']:
                    rtol = 1.e-7
                else:
                    # ceil, floor, and nearest not super well approximated by trapz
                    rtol = 1.e-3
                np.testing.assert_allclose(func_ans, np_ans, rtol=rtol)

                # if xmin > xmax, return the negative.
                if x_min is not None and x_max is not None:
                    neg_func_ans = func.integrate(x_max, x_min)
                    np.testing.assert_allclose(neg_func_ans, -np_ans, rtol=rtol)

                # if xmin == xmax, result should be 0.
                if x_min is not None and x_max is None:
                    zero_func_ans = func.integrate(x_min, x_min)
                    np.testing.assert_allclose(zero_func_ans, 0, atol=1.e-12)


    # Time how long it takes to integrate a really big table
    x = np.linspace(0,2,1000000)
    y = 4. - x**2
    assert np.min(y) >= 0.
    y = np.sqrt(y)
    t0 = time.time()
    ans1 = galsim._LookupTable(x,y,'linear').integrate()
    t1 = time.time()
    ans2 = galsim._LookupTable(x,y,'spline').integrate()
    t2 = time.time()
    ans3 = galsim._LookupTable(x,y,'floor').integrate()
    t3 = time.time()
    ans4 = galsim._LookupTable(x,y,'ceil').integrate()
    t4 = time.time()
    ans5 = galsim._LookupTable(x,y,'nearest').integrate()
    t5 = time.time()
    ans6 = galsim.trapz(y,x)
    t6 = time.time()
    ans7 = np_trapz(y,x)
    t7 = time.time()
    ans8 = galsim.trapz(y,x, interpolant='spline')
    t8 = time.time()
    # Amazingly, LookupTable(linear) is more than 2x faster than np.trapz, and equally accurate.
    # Spline is a bit slower, but about 2x more accurate.
    # Floor and ceil are slightly faster still, but not nearly as accurate.
    print('           integration time       answer - Pi')
    print('linear         %.6f          %.6e'%(t1-t0,ans1-np.pi))
    print('spline         %.6f          %.6e'%(t2-t1,ans2-np.pi))
    print('floor          %.6f          %.6e'%(t3-t2,ans3-np.pi))
    print('ceil           %.6f          %.6e'%(t4-t3,ans4-np.pi))
    print('nearest        %.6f          %.6e'%(t5-t4,ans5-np.pi))
    print('trapz          %.6f          %.6e'%(t6-t5,ans6-np.pi))
    print('np.trapz       %.6f          %.6e'%(t7-t6,ans7-np.pi))
    print('trapz(spline)  %.6f          %.6e'%(t8-t7,ans8-np.pi))
    np.testing.assert_allclose(ans1, np.pi, atol=2.e-9)
    np.testing.assert_allclose(ans2, np.pi, atol=1.e-9)
    np.testing.assert_allclose(ans3, np.pi, atol=3.e-6)
    np.testing.assert_allclose(ans4, np.pi, atol=3.e-6)
    np.testing.assert_allclose(ans5, np.pi, atol=2.e-9)
    np.testing.assert_allclose(ans6, np.pi, atol=2.e-9)
    np.testing.assert_allclose(ans7, np.pi, atol=2.e-9)
    np.testing.assert_allclose(ans8, np.pi, atol=1.e-9)

    # Check errors
    with assert_raises(NotImplementedError):
        galsim.LookupTable([1,2,3,4], [5,5,8,1], x_log=True).integrate()
    with assert_raises(NotImplementedError):
        galsim.LookupTable([1,2,3,4], [5,5,8,1], f_log=True).integrate()
    with assert_raises(NotImplementedError):
        galsim.LookupTable([1,2,3,4], [5,5,8,1], x_log=True, f_log=True).integrate()
    with assert_raises(NotImplementedError):
        galsim.LookupTable([1,2,3,4], [5,5,8,1], interpolant=galsim.Quintic()).integrate()
    with assert_raises(NotImplementedError):
        galsim.LookupTable([1,2,3,4], [5,5,8,1], interpolant=galsim.Linear()).integrate()
    with assert_raises(NotImplementedError):
        galsim.LookupTable([1,2,3,4], [5,5,8,1], interpolant=galsim.Lanczos(5)).integrate()


@timer
def test_integrate_product():
    if np.__version__ >= '2.0':
        np_trapz = np.trapezoid
    else:
        np_trapz = np.trapz

    functions = [
        galsim.LookupTable([0,1,2,3,4], [1,5,5,8,1], interpolant='linear'),
        galsim.LookupTable([0,4], [1,8], interpolant='linear'),
        galsim.LookupTable([0,0.1,0.2,3.5,4], [1,5,5,8,1], interpolant='linear'),
        galsim.LookupTable([0,1,2,3,4], [1,5,5,8,1], interpolant='floor'),
        galsim.LookupTable([0,1,2,3,4], [1,5,5,8,1], interpolant='ceil'),
        galsim.LookupTable([0,1,2,3,4], [1,5,5,8,1], interpolant='nearest'),
        galsim.LookupTable([0,1,2,3,4], [1,5,5,8,1], interpolant='spline'),
        galsim.LookupTable([0,2.1,2.2,2.5,4], [1,5,5,5,1], interpolant='spline'),
    ]

    g_functions = [
        galsim.LookupTable([0,1,2,3,4], [1,5,5,8,1], interpolant='linear'),
        galsim.LookupTable([0,6], [1,8], interpolant='linear'),
        galsim.LookupTable([2,4,6,8], [1,5,8,1], interpolant='spline'),
        galsim.LookupTable([0,2.1,2.2,2.5,4], [1,5,5,5,1], interpolant='spline'),
        lambda x: x**3,
        np.exp,
        lambda x: 17.,
        galsim.LookupTable([5,10], [1,8], interpolant='linear'),
    ]

    for func in functions:
        print('func = ',repr(func))
        for g in g_functions:
            #print('g = ',repr(g))
            for xfact in [1, 1.25, 0.9]:
                #print('xfact = ',xfact)
                for x_min in [None, func.x_min/xfact, func.x_min/xfact+1, func.x_min/xfact-1]:
                    for x_max in [None, func.x_max/xfact, func.x_max/xfact-1, func.x_max/xfact+1]:
                        #print('xmin/xmax = ',x_min,x_max)
                        func_ans = func.integrate_product(g, x_min, x_max, xfact)
                        if x_min is None or x_min < func.x_min:
                            xmin = func.x_min / xfact
                        else:
                            xmin = x_min
                        if x_max is None or x_max > func.x_max:
                            xmax = func.x_max / xfact
                        else:
                            xmax = x_max

                        # Make the version of g that integrate_product approximates it as.
                        if isinstance(g, galsim.LookupTable):
                            if xmin < g.x_min: xmin = g.x_min
                            if xmax > g.x_max: xmax = g.x_max
                            x2 = np.union1d(func.x / xfact, g.x)
                            x2 = x2[(x2>=xmin) & (x2<=xmax)]
                        else:
                            x2 = func.x / xfact
                        if xmin >= xmax:
                            np.testing.assert_allclose(func_ans, 0, atol=1.e-12)
                            continue
                        x2 = np.union1d(x2, [xmin, xmax])
                        gf2 = g(x2)
                        if isinstance(gf2, float): gf2 = gf2 * np.ones(len(x2))
                        g2 = galsim.LookupTable(x2, gf2, 'linear')

                        # Do the integral using trapz with fine x spacing.
                        x = np.linspace(xmin, xmax, 10000)
                        fg = func(x*xfact) * g2(x)
                        np_ans = np_trapz(fg,x)
                        #print('integrate range %s..%s = %s  %s'%(xmin,xmax,func_ans,np_ans))
                        if func.interpolant in ['linear', 'spline']:
                            rtol = 1.e-7
                        else:
                            # ceil, floor, and nearest not super well approximated by trapz
                            rtol = 1.e-3
                        np.testing.assert_allclose(func_ans, np_ans, rtol=rtol)

                        # if xmin > xmax, return the negative.
                        if x_min is not None and x_max is not None:
                            neg_func_ans = func.integrate_product(g, x_max, x_min, xfact)
                            np.testing.assert_allclose(neg_func_ans, -np_ans, rtol=rtol)

                        # if xmin == xmax, result should be 0.
                        if x_min is not None and x_max is None:
                            zero_func_ans = func.integrate_product(g, x_min, x_min, xfact)
                            np.testing.assert_allclose(zero_func_ans, 0, atol=1.e-12)


    # Check errors
    with assert_raises(NotImplementedError):
        galsim.LookupTable([1,2,3,4], [5,5,8,1], x_log=True).integrate_product(g)
    with assert_raises(NotImplementedError):
        galsim.LookupTable([1,2,3,4], [5,5,8,1], f_log=True).integrate_product(g)
    with assert_raises(NotImplementedError):
        galsim.LookupTable([1,2,3,4], [5,5,8,1], x_log=True, f_log=True).integrate_product(g)
    with assert_raises(NotImplementedError):
        galsim.LookupTable([1,2,3,4], [5,5,8,1], interpolant=galsim.Quintic()).integrate_product(g)
    with assert_raises(NotImplementedError):
        galsim.LookupTable([1,2,3,4], [5,5,8,1], interpolant=galsim.Linear()).integrate_product(g)
    with assert_raises(NotImplementedError):
        galsim.LookupTable([1,2,3,4], [5,5,8,1], interpolant=galsim.Lanczos(5)).integrate_product(g)


if __name__ == "__main__":
    runtests(__file__)
